# Generated from src/gsql2rsql/parser/grammar/Cypher.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .CypherParser import CypherParser
else:
    from CypherParser import CypherParser

# This class defines a complete generic visitor for a parse tree produced by CypherParser.

class CypherVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by CypherParser#oC_Cypher.
    def visitOC_Cypher(self, ctx:CypherParser.OC_CypherContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Statement.
    def visitOC_Statement(self, ctx:CypherParser.OC_StatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Query.
    def visitOC_Query(self, ctx:CypherParser.OC_QueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_RegularQuery.
    def visitOC_RegularQuery(self, ctx:CypherParser.OC_RegularQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Union.
    def visitOC_Union(self, ctx:CypherParser.OC_UnionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_SingleQuery.
    def visitOC_SingleQuery(self, ctx:CypherParser.OC_SingleQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_SinglePartQuery.
    def visitOC_SinglePartQuery(self, ctx:CypherParser.OC_SinglePartQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_MultiPartQuery.
    def visitOC_MultiPartQuery(self, ctx:CypherParser.OC_MultiPartQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_UpdatingClause.
    def visitOC_UpdatingClause(self, ctx:CypherParser.OC_UpdatingClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ReadingClause.
    def visitOC_ReadingClause(self, ctx:CypherParser.OC_ReadingClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Match.
    def visitOC_Match(self, ctx:CypherParser.OC_MatchContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Unwind.
    def visitOC_Unwind(self, ctx:CypherParser.OC_UnwindContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Merge.
    def visitOC_Merge(self, ctx:CypherParser.OC_MergeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_MergeAction.
    def visitOC_MergeAction(self, ctx:CypherParser.OC_MergeActionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Create.
    def visitOC_Create(self, ctx:CypherParser.OC_CreateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Set.
    def visitOC_Set(self, ctx:CypherParser.OC_SetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_SetItem.
    def visitOC_SetItem(self, ctx:CypherParser.OC_SetItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Delete.
    def visitOC_Delete(self, ctx:CypherParser.OC_DeleteContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Remove.
    def visitOC_Remove(self, ctx:CypherParser.OC_RemoveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_RemoveItem.
    def visitOC_RemoveItem(self, ctx:CypherParser.OC_RemoveItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_InQueryCall.
    def visitOC_InQueryCall(self, ctx:CypherParser.OC_InQueryCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_StandaloneCall.
    def visitOC_StandaloneCall(self, ctx:CypherParser.OC_StandaloneCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_YieldItems.
    def visitOC_YieldItems(self, ctx:CypherParser.OC_YieldItemsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_YieldItem.
    def visitOC_YieldItem(self, ctx:CypherParser.OC_YieldItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_With.
    def visitOC_With(self, ctx:CypherParser.OC_WithContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Return.
    def visitOC_Return(self, ctx:CypherParser.OC_ReturnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ProjectionBody.
    def visitOC_ProjectionBody(self, ctx:CypherParser.OC_ProjectionBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ProjectionItems.
    def visitOC_ProjectionItems(self, ctx:CypherParser.OC_ProjectionItemsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ProjectionItem.
    def visitOC_ProjectionItem(self, ctx:CypherParser.OC_ProjectionItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Order.
    def visitOC_Order(self, ctx:CypherParser.OC_OrderContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Skip.
    def visitOC_Skip(self, ctx:CypherParser.OC_SkipContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Limit.
    def visitOC_Limit(self, ctx:CypherParser.OC_LimitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_SortItem.
    def visitOC_SortItem(self, ctx:CypherParser.OC_SortItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Where.
    def visitOC_Where(self, ctx:CypherParser.OC_WhereContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Pattern.
    def visitOC_Pattern(self, ctx:CypherParser.OC_PatternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_PatternPart.
    def visitOC_PatternPart(self, ctx:CypherParser.OC_PatternPartContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_AnonymousPatternPart.
    def visitOC_AnonymousPatternPart(self, ctx:CypherParser.OC_AnonymousPatternPartContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_PatternElement.
    def visitOC_PatternElement(self, ctx:CypherParser.OC_PatternElementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_RelationshipsPattern.
    def visitOC_RelationshipsPattern(self, ctx:CypherParser.OC_RelationshipsPatternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_NodePattern.
    def visitOC_NodePattern(self, ctx:CypherParser.OC_NodePatternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_PatternElementChain.
    def visitOC_PatternElementChain(self, ctx:CypherParser.OC_PatternElementChainContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_RelationshipPattern.
    def visitOC_RelationshipPattern(self, ctx:CypherParser.OC_RelationshipPatternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_RelationshipDetail.
    def visitOC_RelationshipDetail(self, ctx:CypherParser.OC_RelationshipDetailContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Properties.
    def visitOC_Properties(self, ctx:CypherParser.OC_PropertiesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_RelationshipTypes.
    def visitOC_RelationshipTypes(self, ctx:CypherParser.OC_RelationshipTypesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_NodeLabels.
    def visitOC_NodeLabels(self, ctx:CypherParser.OC_NodeLabelsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_NodeLabel.
    def visitOC_NodeLabel(self, ctx:CypherParser.OC_NodeLabelContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_RangeLiteral.
    def visitOC_RangeLiteral(self, ctx:CypherParser.OC_RangeLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_LabelName.
    def visitOC_LabelName(self, ctx:CypherParser.OC_LabelNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_RelTypeName.
    def visitOC_RelTypeName(self, ctx:CypherParser.OC_RelTypeNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_PropertyExpression.
    def visitOC_PropertyExpression(self, ctx:CypherParser.OC_PropertyExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Expression.
    def visitOC_Expression(self, ctx:CypherParser.OC_ExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_OrExpression.
    def visitOC_OrExpression(self, ctx:CypherParser.OC_OrExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_XorExpression.
    def visitOC_XorExpression(self, ctx:CypherParser.OC_XorExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_AndExpression.
    def visitOC_AndExpression(self, ctx:CypherParser.OC_AndExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_NotExpression.
    def visitOC_NotExpression(self, ctx:CypherParser.OC_NotExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ComparisonExpression.
    def visitOC_ComparisonExpression(self, ctx:CypherParser.OC_ComparisonExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_PartialComparisonExpression.
    def visitOC_PartialComparisonExpression(self, ctx:CypherParser.OC_PartialComparisonExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_StringListNullPredicateExpression.
    def visitOC_StringListNullPredicateExpression(self, ctx:CypherParser.OC_StringListNullPredicateExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_StringPredicateExpression.
    def visitOC_StringPredicateExpression(self, ctx:CypherParser.OC_StringPredicateExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ListPredicateExpression.
    def visitOC_ListPredicateExpression(self, ctx:CypherParser.OC_ListPredicateExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_NullPredicateExpression.
    def visitOC_NullPredicateExpression(self, ctx:CypherParser.OC_NullPredicateExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_AddOrSubtractExpression.
    def visitOC_AddOrSubtractExpression(self, ctx:CypherParser.OC_AddOrSubtractExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_MultiplyDivideModuloExpression.
    def visitOC_MultiplyDivideModuloExpression(self, ctx:CypherParser.OC_MultiplyDivideModuloExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_PowerOfExpression.
    def visitOC_PowerOfExpression(self, ctx:CypherParser.OC_PowerOfExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_UnaryAddOrSubtractExpression.
    def visitOC_UnaryAddOrSubtractExpression(self, ctx:CypherParser.OC_UnaryAddOrSubtractExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_NonArithmeticOperatorExpression.
    def visitOC_NonArithmeticOperatorExpression(self, ctx:CypherParser.OC_NonArithmeticOperatorExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ListOperatorExpression.
    def visitOC_ListOperatorExpression(self, ctx:CypherParser.OC_ListOperatorExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_PropertyLookup.
    def visitOC_PropertyLookup(self, ctx:CypherParser.OC_PropertyLookupContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Atom.
    def visitOC_Atom(self, ctx:CypherParser.OC_AtomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_CaseExpression.
    def visitOC_CaseExpression(self, ctx:CypherParser.OC_CaseExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_CaseAlternative.
    def visitOC_CaseAlternative(self, ctx:CypherParser.OC_CaseAlternativeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ListComprehension.
    def visitOC_ListComprehension(self, ctx:CypherParser.OC_ListComprehensionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_PatternComprehension.
    def visitOC_PatternComprehension(self, ctx:CypherParser.OC_PatternComprehensionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Quantifier.
    def visitOC_Quantifier(self, ctx:CypherParser.OC_QuantifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Reduce.
    def visitOC_Reduce(self, ctx:CypherParser.OC_ReduceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_FilterExpression.
    def visitOC_FilterExpression(self, ctx:CypherParser.OC_FilterExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_PatternPredicate.
    def visitOC_PatternPredicate(self, ctx:CypherParser.OC_PatternPredicateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ParenthesizedExpression.
    def visitOC_ParenthesizedExpression(self, ctx:CypherParser.OC_ParenthesizedExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_IdInColl.
    def visitOC_IdInColl(self, ctx:CypherParser.OC_IdInCollContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_FunctionInvocation.
    def visitOC_FunctionInvocation(self, ctx:CypherParser.OC_FunctionInvocationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_FunctionName.
    def visitOC_FunctionName(self, ctx:CypherParser.OC_FunctionNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ExistentialSubquery.
    def visitOC_ExistentialSubquery(self, ctx:CypherParser.OC_ExistentialSubqueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ExplicitProcedureInvocation.
    def visitOC_ExplicitProcedureInvocation(self, ctx:CypherParser.OC_ExplicitProcedureInvocationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ImplicitProcedureInvocation.
    def visitOC_ImplicitProcedureInvocation(self, ctx:CypherParser.OC_ImplicitProcedureInvocationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ProcedureResultField.
    def visitOC_ProcedureResultField(self, ctx:CypherParser.OC_ProcedureResultFieldContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ProcedureName.
    def visitOC_ProcedureName(self, ctx:CypherParser.OC_ProcedureNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Namespace.
    def visitOC_Namespace(self, ctx:CypherParser.OC_NamespaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Variable.
    def visitOC_Variable(self, ctx:CypherParser.OC_VariableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Literal.
    def visitOC_Literal(self, ctx:CypherParser.OC_LiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_BooleanLiteral.
    def visitOC_BooleanLiteral(self, ctx:CypherParser.OC_BooleanLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_NumberLiteral.
    def visitOC_NumberLiteral(self, ctx:CypherParser.OC_NumberLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_IntegerLiteral.
    def visitOC_IntegerLiteral(self, ctx:CypherParser.OC_IntegerLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_DoubleLiteral.
    def visitOC_DoubleLiteral(self, ctx:CypherParser.OC_DoubleLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ListLiteral.
    def visitOC_ListLiteral(self, ctx:CypherParser.OC_ListLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_MapLiteral.
    def visitOC_MapLiteral(self, ctx:CypherParser.OC_MapLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_PropertyKeyName.
    def visitOC_PropertyKeyName(self, ctx:CypherParser.OC_PropertyKeyNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Parameter.
    def visitOC_Parameter(self, ctx:CypherParser.OC_ParameterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_SchemaName.
    def visitOC_SchemaName(self, ctx:CypherParser.OC_SchemaNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_ReservedWord.
    def visitOC_ReservedWord(self, ctx:CypherParser.OC_ReservedWordContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_SymbolicName.
    def visitOC_SymbolicName(self, ctx:CypherParser.OC_SymbolicNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_LeftArrowHead.
    def visitOC_LeftArrowHead(self, ctx:CypherParser.OC_LeftArrowHeadContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_RightArrowHead.
    def visitOC_RightArrowHead(self, ctx:CypherParser.OC_RightArrowHeadContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CypherParser#oC_Dash.
    def visitOC_Dash(self, ctx:CypherParser.OC_DashContext):
        return self.visitChildren(ctx)



del CypherParser


"""Resolved column references for the transpiler.

This module provides types for representing fully-resolved column references.
After the resolution pass, every column reference in the query is resolved
to a ResolvedColumnRef that contains all information needed for rendering.

Key types:
- ResolvedColumnRef: A fully resolved reference to a column
- ResolvedExpression: Wrapper for expressions with resolved references
- ColumnRefType: Classification of column reference types
"""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum, auto
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from gsql2rsql.parser.ast import QueryExpression, QueryExpressionProperty


class ColumnRefType(Enum):
    """Classification of column reference types."""

    ENTITY_ID = auto()        # Reference to entity's ID column (e.g., p in RETURN p)
    ENTITY_PROPERTY = auto()  # Reference to entity property (e.g., p.name)
    VALUE_ALIAS = auto()      # Reference to a value alias (e.g., friends after WITH)
    COMPUTED = auto()         # Computed expression result
    LITERAL = auto()          # Literal value (not a reference)


@dataclass
class ResolvedColumnRef:
    """A fully resolved column reference.

    This represents a column reference that has been validated and resolved
    against the symbol table. It contains all information needed to render
    the reference in SQL.

    Attributes:
        original_variable: The variable name as written (e.g., "p")
        original_property: The property name if any (e.g., "name")
        ref_type: Type of reference (entity ID, property, value alias, etc.)
        source_operator_id: ID of the operator that provides this column
        sql_column_name: The SQL column name to use (e.g., "_gsql2rsql_p_name")
        sql_table_alias: The table/subquery alias if known (e.g., "_gsql2rsql_left")
        data_type: Python type of the value (int, str, bool, etc.)
        data_type_name: Human-readable type name (e.g., "STRING", "INTEGER")
        is_nullable: Whether the column can be NULL
        derivation: Human-readable explanation of where this comes from

    Example:
        For `p.name` in:
            MATCH (p:Person) RETURN p.name

        ResolvedColumnRef(
            original_variable="p",
            original_property="name",
            ref_type=ColumnRefType.ENTITY_PROPERTY,
            source_operator_id=1,
            sql_column_name="_gsql2rsql_p_name",
            sql_table_alias=None,
            data_type=str,
            data_type_name="STRING",
            is_nullable=True,
            derivation="p.name from DataSourceOperator(Person)"
        )
    """

    # Original reference
    original_variable: str
    original_property: str | None = None

    # Resolution result
    ref_type: ColumnRefType = ColumnRefType.ENTITY_PROPERTY
    source_operator_id: int = 0
    sql_column_name: str = ""
    sql_table_alias: str | None = None
    is_entity_return: bool = False  # True if this is a bare entity return (RETURN p vs RETURN p.name)

    # Type information
    data_type: type[Any] | None = None
    data_type_name: str | None = None
    is_nullable: bool = True

    # Debugging
    derivation: str = ""

    @property
    def full_sql_reference(self) -> str:
        """Get the full SQL reference including table alias if present.

        Returns:
            SQL reference like "alias.column" or just "column"
        """
        if self.sql_table_alias:
            return f"{self.sql_table_alias}.{self.sql_column_name}"
        return self.sql_column_name

    @property
    def original_text(self) -> str:
        """Get the original Cypher text representation.

        Returns:
            Original reference like "p.name" or "p"
        """
        if self.original_property:
            return f"{self.original_variable}.{self.original_property}"
        return self.original_variable

    def __str__(self) -> str:
        """Human-readable representation."""
        return f"{self.original_text} -> {self.full_sql_reference} ({self.ref_type.name})"


@dataclass
class ResolvedExpression:
    """An expression with all column references resolved.

    This wraps a QueryExpression and provides access to all resolved
    column references within it. Used to pass resolved expressions
    through the rendering pipeline.

    Attributes:
        original_expression: The original AST expression
        column_refs: Map from original reference text to ResolvedColumnRef
        result_type: The computed result type of the expression
        result_sql_name: SQL column name for expression result (if aliased)
    """

    original_expression: QueryExpression
    column_refs: dict[str, ResolvedColumnRef] = field(default_factory=dict)
    result_type: type[Any] | None = None
    result_sql_name: str | None = None

    def get_ref(self, variable: str, property_name: str | None = None) -> ResolvedColumnRef | None:
        """Get the resolved reference for a variable/property.

        Args:
            variable: Variable name (e.g., "p")
            property_name: Optional property name (e.g., "name")

        Returns:
            ResolvedColumnRef if found, None otherwise
        """
        key = f"{variable}.{property_name}" if property_name else variable
        return self.column_refs.get(key)

    def all_refs(self) -> list[ResolvedColumnRef]:
        """Get all resolved column references in this expression.

        Returns:
            List of all ResolvedColumnRef objects
        """
        return list(self.column_refs.values())

    def __str__(self) -> str:
        """Human-readable representation."""
        refs = ", ".join(str(ref) for ref in self.column_refs.values())
        return f"ResolvedExpression({refs})"


@dataclass
class ResolvedProjection:
    """A resolved projection (alias, expression) pair.

    Used in ProjectionOperator to represent a single output column.

    Attributes:
        alias: Output column alias (e.g., "name")
        expression: The resolved expression
        sql_output_name: SQL column name in output (e.g., "_gsql2rsql_name" or just "name")
        is_entity_ref: True if this projects an entire entity (not a property)
        entity_id_column: For entity refs, the ID column name
    """

    alias: str
    expression: ResolvedExpression
    sql_output_name: str = ""
    is_entity_ref: bool = False
    entity_id_column: str | None = None

    def __str__(self) -> str:
        """Human-readable representation."""
        return f"{self.alias}: {self.expression}"


def compute_sql_column_name(
    variable: str,
    property_name: str | None = None,
    prefix: str = "_gsql2rsql_"
) -> str:
    """Compute the deterministic SQL column name for a reference.

    Uses the _gsql2rsql_ prefix to avoid collisions with user identifiers.

    Args:
        variable: Variable name (e.g., "p")
        property_name: Optional property name (e.g., "name")
        prefix: Prefix to use (default: "_gsql2rsql_")

    Returns:
        SQL column name (e.g., "_gsql2rsql_p_name")
    """
    # Clean variable name (remove non-alphanumeric except underscore)
    clean_var = "".join(c if c.isalnum() or c == "_" else "" for c in variable)

    if property_name:
        clean_prop = "".join(c if c.isalnum() or c == "_" else "" for c in property_name)
        return f"{prefix}{clean_var}_{clean_prop}"
    else:
        # Entity reference (ID)
        return f"{prefix}{clean_var}_id"


def compute_sql_table_alias(
    position: str,
    counter: int | None = None,
    prefix: str = "_gsql2rsql_"
) -> str:
    """Compute a deterministic SQL table/subquery alias.

    Args:
        position: Position indicator (e.g., "left", "right", "sq")
        counter: Optional counter for uniqueness
        prefix: Prefix to use (default: "_gsql2rsql_")

    Returns:
        SQL table alias (e.g., "_gsql2rsql_left", "_gsql2rsql_sq_1")
    """
    if counter is not None:
        return f"{prefix}{position}_{counter}"
    return f"{prefix}{position}"


def compute_cte_name(
    purpose: str,
    counter: int,
    prefix: str = "_gsql2rsql_"
) -> str:
    """Compute a deterministic CTE name.

    Args:
        purpose: Purpose of the CTE (e.g., "agg", "recursive")
        counter: Counter for uniqueness
        prefix: Prefix to use (default: "_gsql2rsql_")

    Returns:
        CTE name (e.g., "_gsql2rsql_agg_1")
    """
    return f"{prefix}{purpose}_{counter}"



"""Test: BFS via OpenCypher VLP em PySpark - Social Graph.

Grafo de teste (rede social com multiplos caminhos):

       Alice
      /  |  \\
    Bob Carol Dave
     \\  |   / |
       Evan    Frank
         |
       Grace

Arestas (KNOWS, direcionadas):
  Alice -> Bob, Alice -> Carol, Alice -> Dave
  Bob -> Evan, Carol -> Evan, Dave -> Evan, Dave -> Frank
  Evan -> Grace

Caracteristicas interessantes:
  - Multiplos caminhos para Evan (via Bob, Carol, ou Dave)
  - Dave tem 2 conexoes (Evan e Frank)
  - Frank e "folha" (sem saida)
  - Grace so e alcancavel via Evan
"""

import pytest
from pyspark.sql import SparkSession
from gsql2rsql import GraphContext


@pytest.fixture(scope="module")
def spark():
    """Create SparkSession for tests."""
    spark = (
        SparkSession.builder
        .appName("BFS_SocialGraph_Test")
        .master("local[2]")
        .config("spark.sql.shuffle.partitions", "2")
        .config("spark.driver.memory", "1g")
        .getOrCreate()
    )
    yield spark
    spark.stop()


@pytest.fixture(scope="module")
def social_graph(spark):
    """Create social network graph data."""
    # Nodes: people with different roles
    nodes_data = [
        ("Alice", "Person", "founder"),
        ("Bob", "Person", "employee"),
        ("Carol", "Person", "employee"),
        ("Dave", "Person", "manager"),
        ("Evan", "Person", "intern"),
        ("Frank", "Person", "contractor"),
        ("Grace", "Person", "intern"),
    ]
    nodes_df = spark.createDataFrame(
        nodes_data, ["node_id", "node_type", "role"]
    )
    nodes_df.createOrReplaceTempView("people")

    # Edges: KNOWS relationships
    edges_data = [
        ("Alice", "Bob", "KNOWS"),
        ("Alice", "Carol", "KNOWS"),
        ("Alice", "Dave", "KNOWS"),
        ("Bob", "Evan", "KNOWS"),
        ("Carol", "Evan", "KNOWS"),
        ("Dave", "Evan", "KNOWS"),
        ("Dave", "Frank", "KNOWS"),
        ("Evan", "Grace", "KNOWS"),
    ]
    edges_df = spark.createDataFrame(
        edges_data, ["src", "dst", "relationship_type"]
    )
    edges_df.createOrReplaceTempView("connections")

    return {"nodes": nodes_df, "edges": edges_df}


@pytest.fixture(scope="module")
def graph_context(spark, social_graph):
    """Create GraphContext for social network."""
    graph = GraphContext(
        spark=spark,
        nodes_table="people",
        edges_table="connections",
        node_type_col="node_type",
        node_id_col="node_id",
        extra_node_attrs={"role": str},
    )
    graph.set_types(
        node_types=["Person"],
        edge_types=["KNOWS"],
    )
    return graph


class TestBFSSocialGraph:
    """BFS tests on social network graph."""

    def test_bfs_depth_1_direct_friends(self, spark, graph_context):
        """BFS depth 1: Alice -> {Bob, Carol, Dave}."""
        query = """
        MATCH (person)-[:KNOWS*1..1]->(friend)
        WHERE person.node_id = 'Alice'
        RETURN person.node_id AS person, friend.node_id AS friend
        ORDER BY friend
        """

        sql = graph_context.transpile(query)
        print("\n=== SQL para BFS depth 1 (amigos diretos) ===")
        print(sql)

        result = spark.sql(sql)
        rows = result.collect()

        expected = 3
        assert len(rows) == expected, f"Alice tem {expected} amigos, got {len(rows)}"

        friends = {row["friend"] for row in rows}
        assert friends == {"Bob", "Carol", "Dave"}, f"Amigos: {friends}"

        print(f"BFS depth 1: Alice -> {friends}")

    def test_bfs_depth_2_friends_of_friends(self, spark, graph_context):
        """BFS depth 2: Alice pode alcancar Evan, Frank via depth 2."""
        query = """
        MATCH path = (person)-[:KNOWS*2..2]->(fof)
        WHERE person.node_id = 'Alice'
        RETURN person.node_id AS person, fof.node_id AS friend_of_friend
        """

        sql = graph_context.transpile(query)
        print("\n=== SQL para BFS depth 2 (amigos de amigos) ===")
        print(sql)

        result = spark.sql(sql)
        rows = result.collect()

        # Alice -> Bob/Carol/Dave -> Evan (3 paths), Dave -> Frank (1 path)
        # Total: 4 paths at depth 2
        expected = 4
        assert len(rows) == expected, f"Esperado {expected} caminhos, got {len(rows)}"

        fof_names = {row["friend_of_friend"] for row in rows}
        assert "Evan" in fof_names, "Evan deve ser alcancavel em depth 2"
        assert "Frank" in fof_names, "Frank deve ser alcancavel em depth 2"

        print(f"BFS depth 2: {fof_names}")

    def test_bfs_multiple_paths_to_same_node(self, spark, graph_context):
        """Verificar multiplos caminhos para Evan."""
        query = """
        MATCH path = (alice)-[:KNOWS*2..2]->(evan)
        WHERE alice.node_id = 'Alice' AND evan.node_id = 'Evan'
        RETURN nodes(path) AS path_nodes
        """

        sql = graph_context.transpile(query)
        print("\n=== SQL para multiplos caminhos para Evan ===")
        print(sql)

        result = spark.sql(sql)
        rows = result.collect()

        # 3 caminhos: Alice->Bob->Evan, Alice->Carol->Evan, Alice->Dave->Evan
        expected = 3
        assert len(rows) == expected, f"Esperado {expected} caminhos, got {len(rows)}"

        paths = [tuple(row["path_nodes"]) for row in rows]
        expected_intermediates = {"Bob", "Carol", "Dave"}
        actual_intermediates = {path[1] for path in paths}

        assert actual_intermediates == expected_intermediates, \
            f"Intermediarios: {actual_intermediates}"

        print(f"Multiplos caminhos para Evan: {paths}")

    def test_bfs_depth_3_reaches_grace(self, spark, graph_context):
        """BFS depth 3: Alice -> ... -> Evan -> Grace."""
        query = """
        MATCH path = (alice)-[:KNOWS*3..3]->(grace)
        WHERE alice.node_id = 'Alice' AND grace.node_id = 'Grace'
        RETURN nodes(path) AS path_nodes, length(path) AS depth
        """

        sql = graph_context.transpile(query)
        print("\n=== SQL para BFS depth 3 (alcancar Grace) ===")
        print(sql)

        result = spark.sql(sql)
        rows = result.collect()

        # 3 caminhos de depth 3 para Grace
        expected = 3
        assert len(rows) == expected, f"Esperado {expected} caminhos, got {len(rows)}"

        for row in rows:
            path = row["path_nodes"]
            assert path[0] == "Alice", "Deve comecar em Alice"
            assert path[-1] == "Grace", "Deve terminar em Grace"
            assert path[-2] == "Evan", "Penultimo deve ser Evan"
            assert row["depth"] == 3, "Depth deve ser 3"

        print(f"BFS depth 3: {len(rows)} caminhos ate Grace")

    def test_bfs_all_reachable_from_alice(self, spark, graph_context):
        """BFS depth 1-3: todos os nos alcancaveis de Alice."""
        query = """
        MATCH path = (alice)-[:KNOWS*1..3]->(reachable)
        WHERE alice.node_id = 'Alice'
        RETURN DISTINCT reachable.node_id AS reachable_node
        ORDER BY reachable_node
        """

        sql = graph_context.transpile(query)
        print("\n=== SQL para todos alcancaveis de Alice ===")
        print(sql)

        result = spark.sql(sql)
        rows = result.collect()

        reachable = {row["reachable_node"] for row in rows}
        expected = {"Bob", "Carol", "Dave", "Evan", "Frank", "Grace"}

        assert reachable == expected, f"Alcancaveis: {reachable}"

        print(f"Todos alcancaveis de Alice: {reachable}")

    def test_bfs_from_dave_two_branches(self, spark, graph_context):
        """BFS de Dave: tem 2 caminhos diretos (Evan e Frank)."""
        query = """
        MATCH (dave)-[:KNOWS*1..1]->(direct)
        WHERE dave.node_id = 'Dave'
        RETURN dave.node_id AS start, direct.node_id AS direct_connection
        ORDER BY direct_connection
        """

        sql = graph_context.transpile(query)
        print("\n=== SQL para BFS de Dave ===")
        print(sql)

        result = spark.sql(sql)
        rows = result.collect()

        expected = 2
        assert len(rows) == expected, f"Dave tem {expected} conexoes, got {len(rows)}"

        connections = {row["direct_connection"] for row in rows}
        assert connections == {"Evan", "Frank"}, f"Conexoes: {connections}"

        print(f"BFS de Dave: {connections}")

    def test_bfs_frank_is_leaf(self, spark, graph_context):
        """BFS de Frank: e folha, nao tem conexoes de saida."""
        query = """
        MATCH (frank)-[:KNOWS*1..1]->(anyone)
        WHERE frank.node_id = 'Frank'
        RETURN frank.node_id AS start, anyone.node_id AS connection
        """

        sql = graph_context.transpile(query)
        print("\n=== SQL para BFS de Frank (folha) ===")
        print(sql)

        result = spark.sql(sql)
        rows = result.collect()

        assert len(rows) == 0, f"Frank e folha. Encontrado {len(rows)}"

        print("BFS de Frank: nenhuma conexao (correto - e folha)")

    def test_bfs_with_role_attribute(self, spark, graph_context):
        """BFS com atributo role dos nos."""
        query = """
        MATCH path = (alice)-[:KNOWS*1..2]->(contact)
        WHERE alice.node_id = 'Alice'
        RETURN contact.node_id AS name, contact.role AS role,
               length(path) AS distance
        ORDER BY distance, name
        """

        sql = graph_context.transpile(query)
        print("\n=== SQL para BFS com atributo role ===")
        print(sql)

        result = spark.sql(sql)
        rows = result.collect()

        # Verificar roles
        role_map = {row["name"]: row["role"] for row in rows}

        # Depth 1: Bob, Carol, Dave
        assert role_map.get("Bob") == "employee"
        assert role_map.get("Carol") == "employee"
        assert role_map.get("Dave") == "manager"

        # Depth 2: Evan (intern), Frank (contractor)
        evan_rows = [r for r in rows if r["name"] == "Evan"]
        frank_rows = [r for r in rows if r["name"] == "Frank"]

        assert all(r["role"] == "intern" for r in evan_rows)
        assert all(r["role"] == "contractor" for r in frank_rows)

        print(f"BFS com roles: {role_map}")

    def test_bfs_count_paths_by_depth(self, spark, graph_context):
        """Contar caminhos por profundidade."""
        query = """
        MATCH path = (alice)-[:KNOWS*1..3]->(anyone)
        WHERE alice.node_id = 'Alice'
        RETURN length(path) AS depth, COUNT(*) AS path_count
        ORDER BY depth
        """

        sql = graph_context.transpile(query)
        print("\n=== SQL para contar caminhos por depth ===")
        print(sql)

        result = spark.sql(sql)
        rows = result.collect()

        depth_counts = {row["depth"]: row["path_count"] for row in rows}

        # Depth 1: 3 caminhos (Bob, Carol, Dave)
        # Depth 2: 4 caminhos (Evan x3, Frank x1)
        # Depth 3: 3 caminhos (Grace x3, via diferentes intermediarios)
        assert depth_counts.get(1) == 3, f"Depth 1: {depth_counts.get(1)}"
        assert depth_counts.get(2) == 4, f"Depth 2: {depth_counts.get(2)}"
        assert depth_counts.get(3) == 3, f"Depth 3: {depth_counts.get(3)}"

        print(f"Caminhos por depth: {depth_counts}")



"""Binary and unary operators for openCypher expressions."""

from enum import Enum, auto
from typing import NamedTuple


class BinaryOperator(Enum):
    """Binary operators supported by the transpiler."""

    INVALID = auto()

    # Numerical operators
    PLUS = auto()
    MINUS = auto()
    MULTIPLY = auto()
    DIVIDE = auto()
    MODULO = auto()
    EXPONENTIATION = auto()

    # Logical operators
    AND = auto()
    OR = auto()
    XOR = auto()

    # Comparison operators
    LT = auto()
    LEQ = auto()
    GT = auto()
    GEQ = auto()
    EQ = auto()
    NEQ = auto()
    REGMATCH = auto()
    IN = auto()


class BinaryOperatorType(Enum):
    """Classification of binary operators by their type semantics."""

    INVALID = auto()
    VALUE = auto()  # Takes value type (string/numeric) and outputs value type
    LOGICAL = auto()  # Takes logical type (bool) and outputs logical type
    COMPARISON = auto()  # Takes value type (string/numeric) and outputs logical type


class BinaryOperatorInfo(NamedTuple):
    """Information about a binary operator."""

    name: BinaryOperator
    operator_type: BinaryOperatorType

    def __str__(self) -> str:
        return self.name.name


# Mapping from operator symbols to their info
OPERATORS: dict[str, BinaryOperatorInfo] = {
    "+": BinaryOperatorInfo(BinaryOperator.PLUS, BinaryOperatorType.VALUE),
    "-": BinaryOperatorInfo(BinaryOperator.MINUS, BinaryOperatorType.VALUE),
    "*": BinaryOperatorInfo(BinaryOperator.MULTIPLY, BinaryOperatorType.VALUE),
    "/": BinaryOperatorInfo(BinaryOperator.DIVIDE, BinaryOperatorType.VALUE),
    "%": BinaryOperatorInfo(BinaryOperator.MODULO, BinaryOperatorType.VALUE),
    "^": BinaryOperatorInfo(BinaryOperator.EXPONENTIATION, BinaryOperatorType.VALUE),
    "<>": BinaryOperatorInfo(BinaryOperator.NEQ, BinaryOperatorType.COMPARISON),
    "=": BinaryOperatorInfo(BinaryOperator.EQ, BinaryOperatorType.COMPARISON),
    "<": BinaryOperatorInfo(BinaryOperator.LT, BinaryOperatorType.COMPARISON),
    ">": BinaryOperatorInfo(BinaryOperator.GT, BinaryOperatorType.COMPARISON),
    "<=": BinaryOperatorInfo(BinaryOperator.LEQ, BinaryOperatorType.COMPARISON),
    ">=": BinaryOperatorInfo(BinaryOperator.GEQ, BinaryOperatorType.COMPARISON),
    "=~": BinaryOperatorInfo(BinaryOperator.REGMATCH, BinaryOperatorType.COMPARISON),
    "in": BinaryOperatorInfo(BinaryOperator.IN, BinaryOperatorType.COMPARISON),
    "and": BinaryOperatorInfo(BinaryOperator.AND, BinaryOperatorType.LOGICAL),
    "or": BinaryOperatorInfo(BinaryOperator.OR, BinaryOperatorType.LOGICAL),
    "xor": BinaryOperatorInfo(BinaryOperator.XOR, BinaryOperatorType.LOGICAL),
}


def try_get_operator(op: str) -> BinaryOperatorInfo | None:
    """
    Try to get operator info for the given operator symbol.

    Args:
        op: The operator symbol (e.g., "+", "and").

    Returns:
        BinaryOperatorInfo if found, None otherwise.
    """
    return OPERATORS.get(op.lower() if op else "")


class AggregationFunction(Enum):
    """Aggregation functions supported by the transpiler."""

    INVALID = auto()
    NONE = auto()
    SUM = auto()
    AVG = auto()
    COUNT = auto()
    MAX = auto()
    MIN = auto()
    FIRST = auto()
    LAST = auto()
    PERCENTILE_CONT = auto()
    PERCENTILE_DISC = auto()
    STDEV = auto()
    STDEVP = auto()
    COLLECT = auto()


def try_parse_aggregation_function(function_name: str) -> AggregationFunction | None:
    """
    Try to parse a function name into an aggregation function.

    Args:
        function_name: The function name to parse.

    Returns:
        AggregationFunction if it's an aggregation function, None otherwise.
    """
    mapping = {
        "avg": AggregationFunction.AVG,
        "sum": AggregationFunction.SUM,
        "count": AggregationFunction.COUNT,
        "max": AggregationFunction.MAX,
        "min": AggregationFunction.MIN,
        "first": AggregationFunction.FIRST,
        "last": AggregationFunction.LAST,
        "percentilecont": AggregationFunction.PERCENTILE_CONT,
        "percentiledisc": AggregationFunction.PERCENTILE_DISC,
        "stdev": AggregationFunction.STDEV,
        "stddev": AggregationFunction.STDEV,  # Alias with double D
        "stdevp": AggregationFunction.STDEVP,
        "stddevp": AggregationFunction.STDEVP,  # Alias with double D
        "collect": AggregationFunction.COLLECT,
    }
    return mapping.get(function_name.lower())


class ListPredicateType(Enum):
    """Types of list predicate expressions (quantifiers)."""

    ALL = auto()  # ALL(x IN list WHERE pred)
    ANY = auto()  # ANY(x IN list WHERE pred)
    NONE = auto()  # NONE(x IN list WHERE pred)
    SINGLE = auto()  # SINGLE(x IN list WHERE pred)


class Function(Enum):
    """Functions supported by the transpiler."""

    INVALID = auto()

    # Unary operators
    POSITIVE = auto()
    NEGATIVE = auto()
    NOT = auto()

    # Type conversion functions
    TO_FLOAT = auto()
    TO_STRING = auto()
    TO_BOOLEAN = auto()
    TO_INTEGER = auto()
    TO_LONG = auto()
    TO_DOUBLE = auto()

    # String functions
    STRING_STARTS_WITH = auto()
    STRING_ENDS_WITH = auto()
    STRING_CONTAINS = auto()
    STRING_LEFT = auto()
    STRING_RIGHT = auto()
    STRING_TRIM = auto()
    STRING_LTRIM = auto()
    STRING_RTRIM = auto()
    STRING_TO_UPPER = auto()
    STRING_TO_LOWER = auto()
    STRING_SIZE = auto()

    # Misc functions
    IS_NULL = auto()
    IS_NOT_NULL = auto()

    # Null handling functions
    COALESCE = auto()

    # List/Array functions
    RANGE = auto()  # Cypher: RANGE(start, end[, step]) -> Databricks: SEQUENCE
    SIZE = auto()  # Cypher: SIZE(list) -> Databricks: SIZE(list)

    # Path functions
    NODES = auto()  # nodes(path) -> returns array of node IDs
    RELATIONSHIPS = auto()  # relationships(path) -> returns array of edge structs
    LENGTH = auto()  # length(path) -> number of relationships (edges) in path

    # Math functions
    ABS = auto()
    CEIL = auto()
    FLOOR = auto()
    ROUND = auto()
    SQRT = auto()
    SIGN = auto()
    LOG = auto()  # Natural log
    LOG10 = auto()
    EXP = auto()
    SIN = auto()
    COS = auto()
    TAN = auto()
    ASIN = auto()
    ACOS = auto()
    ATAN = auto()
    ATAN2 = auto()
    DEGREES = auto()
    RADIANS = auto()
    RAND = auto()
    PI = auto()
    E = auto()

    # Date/Time functions
    DATE = auto()  # date() or date({...})
    DATETIME = auto()  # datetime() or datetime({...})
    LOCALDATETIME = auto()  # localdatetime()
    TIME = auto()  # time()
    LOCALTIME = auto()  # localtime()
    DURATION = auto()  # duration({...})
    DURATION_BETWEEN = auto()  # duration.between(d1, d2)

    # Date/Time component extraction
    DATE_YEAR = auto()  # date.year or datetime.year
    DATE_MONTH = auto()  # date.month or datetime.month
    DATE_DAY = auto()  # date.day or datetime.day
    DATE_HOUR = auto()  # datetime.hour or time.hour
    DATE_MINUTE = auto()  # datetime.minute or time.minute
    DATE_SECOND = auto()  # datetime.second or time.second
    DATE_WEEK = auto()  # date.week
    DATE_DAYOFWEEK = auto()  # date.dayOfWeek
    DATE_QUARTER = auto()  # date.quarter

    # Date/Time arithmetic and manipulation
    DATE_TRUNCATE = auto()  # date.truncate('unit', d)
    DATE_ADD = auto()  # date + duration -> DATE_ADD
    DATE_SUB = auto()  # date - duration -> DATE_SUB


class FunctionInfo(NamedTuple):
    """Information about a function."""

    function_name: Function
    required_parameters: int
    optional_parameters: int = 0


# Mapping from function names to their info
FUNCTIONS: dict[str, FunctionInfo] = {
    "+": FunctionInfo(Function.POSITIVE, 1),
    "-": FunctionInfo(Function.NEGATIVE, 1),
    "not": FunctionInfo(Function.NOT, 1),
    "tofloat": FunctionInfo(Function.TO_FLOAT, 1),
    "todouble": FunctionInfo(Function.TO_DOUBLE, 1),
    "tostring": FunctionInfo(Function.TO_STRING, 1),
    "toboolean": FunctionInfo(Function.TO_BOOLEAN, 1),
    "tointeger": FunctionInfo(Function.TO_INTEGER, 1),
    "tolong": FunctionInfo(Function.TO_LONG, 1),
    "startswith": FunctionInfo(Function.STRING_STARTS_WITH, 2),
    "endswith": FunctionInfo(Function.STRING_ENDS_WITH, 2),
    "contains": FunctionInfo(Function.STRING_CONTAINS, 2),
    "left": FunctionInfo(Function.STRING_LEFT, 2),
    "right": FunctionInfo(Function.STRING_RIGHT, 2),
    "trim": FunctionInfo(Function.STRING_TRIM, 1),
    "ltrim": FunctionInfo(Function.STRING_LTRIM, 1),
    "rtrim": FunctionInfo(Function.STRING_RTRIM, 1),
    "toupper": FunctionInfo(Function.STRING_TO_UPPER, 1),
    "tolower": FunctionInfo(Function.STRING_TO_LOWER, 1),
    # Null handling - COALESCE takes 1+ args (variadic)
    "coalesce": FunctionInfo(Function.COALESCE, 1, 99),
    # List/Array functions
    "range": FunctionInfo(Function.RANGE, 2, 1),  # RANGE(start, end[, step])
    "size": FunctionInfo(Function.SIZE, 1),  # SIZE works for both strings and arrays
    "length": FunctionInfo(Function.LENGTH, 1),  # LENGTH(path) = number of edges (hops)
    # Path functions
    "nodes": FunctionInfo(Function.NODES, 1),  # nodes(path) -> array of node IDs
    "relationships": FunctionInfo(Function.RELATIONSHIPS, 1),  # rels(path) -> array of edges
    "rels": FunctionInfo(Function.RELATIONSHIPS, 1),  # alias for relationships
    # Math functions
    "abs": FunctionInfo(Function.ABS, 1),
    "ceil": FunctionInfo(Function.CEIL, 1),
    "ceiling": FunctionInfo(Function.CEIL, 1),  # alias
    "floor": FunctionInfo(Function.FLOOR, 1),
    "round": FunctionInfo(Function.ROUND, 1, 1),  # ROUND(x) or ROUND(x, precision)
    "sqrt": FunctionInfo(Function.SQRT, 1),
    "sign": FunctionInfo(Function.SIGN, 1),
    "log": FunctionInfo(Function.LOG, 1),
    "ln": FunctionInfo(Function.LOG, 1),  # alias
    "log10": FunctionInfo(Function.LOG10, 1),
    "exp": FunctionInfo(Function.EXP, 1),
    "sin": FunctionInfo(Function.SIN, 1),
    "cos": FunctionInfo(Function.COS, 1),
    "tan": FunctionInfo(Function.TAN, 1),
    "asin": FunctionInfo(Function.ASIN, 1),
    "acos": FunctionInfo(Function.ACOS, 1),
    "atan": FunctionInfo(Function.ATAN, 1),
    "atan2": FunctionInfo(Function.ATAN2, 2),
    "degrees": FunctionInfo(Function.DEGREES, 1),
    "radians": FunctionInfo(Function.RADIANS, 1),
    "rand": FunctionInfo(Function.RAND, 0),
    "random": FunctionInfo(Function.RAND, 0),  # alias
    "pi": FunctionInfo(Function.PI, 0),
    "e": FunctionInfo(Function.E, 0),
    # Date/Time functions
    "date": FunctionInfo(Function.DATE, 0, 1),  # date() or date({...})
    "datetime": FunctionInfo(Function.DATETIME, 0, 1),  # datetime() or datetime({...})
    "timestamp": FunctionInfo(Function.DATETIME, 0, 1),  # timestamp() -> same as datetime()
    "localdatetime": FunctionInfo(Function.LOCALDATETIME, 0, 1),
    "time": FunctionInfo(Function.TIME, 0, 1),  # time() or time({...})
    "localtime": FunctionInfo(Function.LOCALTIME, 0, 1),
    "duration": FunctionInfo(Function.DURATION, 1),  # duration({...})
    "duration.between": FunctionInfo(Function.DURATION_BETWEEN, 2),
    # Date component extraction (used as methods on date/datetime values)
    "year": FunctionInfo(Function.DATE_YEAR, 1),
    "month": FunctionInfo(Function.DATE_MONTH, 1),
    "day": FunctionInfo(Function.DATE_DAY, 1),
    "hour": FunctionInfo(Function.DATE_HOUR, 1),
    "minute": FunctionInfo(Function.DATE_MINUTE, 1),
    "second": FunctionInfo(Function.DATE_SECOND, 1),
    "week": FunctionInfo(Function.DATE_WEEK, 1),
    "dayofweek": FunctionInfo(Function.DATE_DAYOFWEEK, 1),
    "quarter": FunctionInfo(Function.DATE_QUARTER, 1),
    # Date truncation
    "date.truncate": FunctionInfo(Function.DATE_TRUNCATE, 2),
}


def try_get_function(function_name: str) -> FunctionInfo | None:
    """
    Try to get function info for the given function name.

    Args:
        function_name: The function name.

    Returns:
        FunctionInfo if found, None otherwise.
    """
    return FUNCTIONS.get(function_name.lower() if function_name else "")



"""Tests for the graph schema."""

import pytest

from gsql2rsql.common.schema import (
    EdgeSchema,
    EntityProperty,
    NodeSchema,
    SimpleGraphSchemaProvider,
)


class TestEntityProperty:
    """Tests for EntityProperty."""

    def test_create_property(self) -> None:
        """Test property creation."""
        prop = EntityProperty(property_name="name", data_type=str)
        assert prop.property_name == "name"
        assert prop.data_type == str

    def test_empty_name_raises(self) -> None:
        """Test that empty property name raises."""
        with pytest.raises(ValueError):
            EntityProperty(property_name="", data_type=str)


class TestNodeSchema:
    """Tests for NodeSchema."""

    def test_create_node_schema(self) -> None:
        """Test node schema creation."""
        schema = NodeSchema(
            name="Person",
            properties=[
                EntityProperty(property_name="name", data_type=str),
                EntityProperty(property_name="age", data_type=int),
            ],
        )
        assert schema.name == "Person"
        assert schema.id == "Person"
        assert len(schema.properties) == 2


class TestEdgeSchema:
    """Tests for EdgeSchema."""

    def test_create_edge_schema(self) -> None:
        """Test edge schema creation."""
        schema = EdgeSchema(
            name="KNOWS",
            source_node_id="Person",
            sink_node_id="Person",
        )
        assert schema.name == "KNOWS"
        assert schema.source_node_id == "Person"
        assert schema.sink_node_id == "Person"

    def test_edge_id(self) -> None:
        """Test edge ID generation."""
        schema = EdgeSchema(
            name="ACTED_IN",
            source_node_id="Person",
            sink_node_id="Movie",
        )
        expected_id = "Person@ACTED_IN@Movie"
        assert schema.id == expected_id

    def test_get_edge_id_class_method(self) -> None:
        """Test static edge ID generation."""
        edge_id = EdgeSchema.get_edge_id("KNOWS", "Person", "Person")
        assert edge_id == "Person@KNOWS@Person"


class TestSimpleGraphSchemaProvider:
    """Tests for SimpleGraphSchemaProvider."""

    def test_add_and_get_node(self) -> None:
        """Test adding and retrieving node schema."""
        provider = SimpleGraphSchemaProvider()

        schema = NodeSchema(name="Person")
        provider.add_node(schema)

        result = provider.get_node_definition("Person")
        assert result is not None
        assert result.name == "Person"

    def test_get_nonexistent_node(self) -> None:
        """Test getting non-existent node returns None."""
        provider = SimpleGraphSchemaProvider()
        result = provider.get_node_definition("Unknown")
        assert result is None

    def test_add_and_get_edge(self) -> None:
        """Test adding and retrieving edge schema."""
        provider = SimpleGraphSchemaProvider()

        schema = EdgeSchema(
            name="KNOWS",
            source_node_id="Person",
            sink_node_id="Person",
        )
        provider.add_edge(schema)

        result = provider.get_edge_definition("KNOWS", "Person", "Person")
        assert result is not None
        assert result.name == "KNOWS"

    def test_get_nonexistent_edge(self) -> None:
        """Test getting non-existent edge returns None."""
        provider = SimpleGraphSchemaProvider()
        result = provider.get_edge_definition("UNKNOWN", "A", "B")
        assert result is None



"""Test 26: UNWIND cartesian product (multiple UNWINDs).

Validates that multiple UNWIND clauses produce a cartesian product,
with each UNWIND adding another EXPLODE TVF in the FROM clause.

Cypher:
    UNWIND list1 AS x
    UNWIND list2 AS y
    → produces cartesian product of list1 × list2

Databricks SQL:
    FROM source,
    EXPLODE(list1) AS _exploded1(x),
    EXPLODE(list2) AS _exploded2(y)

Fraud Use Cases:
- Cross-reference multiple watchlists
- Generate all transaction pairs for comparison
- Expand multi-dimensional risk factors
"""



from gsql2rsql import OpenCypherParser, LogicalPlan, SQLRenderer
from gsql2rsql.common.schema import (
    SimpleGraphSchemaProvider,
    NodeSchema,
    EdgeSchema,
    EntityProperty,
)
from gsql2rsql.renderer.schema_provider import (
    SimpleSQLSchemaProvider,
    SQLTableDescriptor,
)

from tests.utils.sql_test_utils import assert_sql_equal, load_expected_sql


class TestUnwindCartesian:
    """Test UNWIND cartesian product transpilation."""

    TEST_ID = "26"
    TEST_NAME = "unwind_cartesian"

    def setup_method(self) -> None:
        """Set up test fixtures."""
        self.graph_schema = SimpleGraphSchemaProvider()
        self.graph_schema.add_node(
            NodeSchema(
                name="Account",
                properties=[
                    EntityProperty("id", str),
                    EntityProperty("name", str),
                    EntityProperty("tags", list),
                    EntityProperty("transactionIds", list),
                    EntityProperty("riskFactors", list),
                ],
                node_id_property=EntityProperty("id", str),
            )
        )

        self.sql_schema = SimpleSQLSchemaProvider()
        self.sql_schema.add_node(
            NodeSchema(
                name="Account",
                properties=[
                    EntityProperty("id", str),
                    EntityProperty("name", str),
                    EntityProperty("tags", list),
                    EntityProperty("transactionIds", list),
                    EntityProperty("riskFactors", list),
                ],
                node_id_property=EntityProperty("id", str),
            ),
            SQLTableDescriptor(table_name="graph.Account"),
        )

    def _transpile(self, cypher: str) -> str:
        """Helper to transpile a Cypher query."""
        parser = OpenCypherParser()
        ast = parser.parse(cypher)
        plan = LogicalPlan.process_query_tree(ast, self.graph_schema)
        plan.resolve(original_query=cypher)
        renderer = SQLRenderer(db_schema_provider=self.sql_schema)
        return renderer.render_plan(plan)

    def test_golden_file_match(self) -> None:
        """Test that transpiled SQL matches golden file."""
        cypher = """
        MATCH (a:Account)
        UNWIND a.tags AS tag
        UNWIND a.transactionIds AS txId
        RETURN a.id, tag, txId
        """
        actual_sql = self._transpile(cypher)

        expected_sql = load_expected_sql(self.TEST_ID, self.TEST_NAME)
        if expected_sql is None:
            from tests.utils.sql_test_utils import EXPECTED_DIR

            raise AssertionError(
                f"No expected SQL file found at "
                f"{EXPECTED_DIR}/{self.TEST_ID}_{self.TEST_NAME}.sql. "
                "Create it first using: make dump-sql ID=26 NAME=unwind_cartesian"
            )

        assert_sql_equal(expected_sql, actual_sql, self.TEST_ID, self.TEST_NAME)

    def test_two_unwinds_produce_two_explodes(self) -> None:
        """Test that two UNWINDs produce two EXPLODE functions."""
        cypher = """
        MATCH (a:Account)
        UNWIND a.tags AS tag
        UNWIND a.transactionIds AS txId
        RETURN a.id, tag, txId
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        explode_count = sql_upper.count("EXPLODE(")
        assert explode_count >= 2, \
            f"Two UNWINDs should produce >= 2 EXPLODEs, found {explode_count}"

    def test_three_unwinds_produce_three_explodes(self) -> None:
        """Test that three UNWINDs produce three EXPLODE functions."""
        cypher = """
        MATCH (a:Account)
        UNWIND a.tags AS tag
        UNWIND a.transactionIds AS txId
        UNWIND a.riskFactors AS risk
        RETURN a.id, tag, txId, risk
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        explode_count = sql_upper.count("EXPLODE(")
        assert explode_count >= 3, \
            f"Three UNWINDs should produce >= 3 EXPLODEs, found {explode_count}"

    def test_cartesian_with_literal_arrays(self) -> None:
        """Test cartesian product with literal arrays."""
        cypher = """
        MATCH (a:Account)
        UNWIND [1, 2] AS x
        UNWIND ['a', 'b'] AS y
        RETURN a.id, x, y
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        explode_count = sql_upper.count("EXPLODE(")
        assert explode_count >= 2
        # Should contain all literal values
        assert "1" in sql and "2" in sql
        assert "'A'" in sql_upper or "'a'" in sql.lower()

    def test_cartesian_preserves_source_columns(self) -> None:
        """Test that cartesian product preserves source columns."""
        cypher = """
        MATCH (a:Account)
        UNWIND a.tags AS tag
        UNWIND a.transactionIds AS txId
        RETURN a.id, a.name, tag, txId
        """
        sql = self._transpile(cypher)

        sql_lower = sql.lower()
        assert "id" in sql_lower, "Should preserve id column"
        assert "name" in sql_lower, "Should preserve name column"

    def test_cartesian_no_lateral_keyword(self) -> None:
        """Test that cartesian uses TVF syntax without deprecated LATERAL."""
        cypher = """
        MATCH (a:Account)
        UNWIND a.tags AS tag
        UNWIND a.transactionIds AS txId
        RETURN a.id, tag, txId
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "LATERAL" not in sql_upper, \
            "Should NOT use deprecated LATERAL keyword"

    def test_fraud_cross_reference_watchlists(self) -> None:
        """Test fraud scenario: cross-reference multiple arrays.

        Pattern: Compare each tag with each risk factor.
        """
        cypher = """
        MATCH (a:Account)
        UNWIND a.tags AS tag
        UNWIND a.riskFactors AS risk
        RETURN a.id, tag, risk
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert sql_upper.count("EXPLODE(") >= 2

    def test_cartesian_with_filter_after(self) -> None:
        """Test cartesian product with filter on combined results."""
        cypher = """
        MATCH (a:Account)
        UNWIND a.tags AS tag
        UNWIND a.riskFactors AS risk
        WITH a, tag, risk
        WHERE tag = risk
        RETURN a.id, tag
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert sql_upper.count("EXPLODE(") >= 2
        # Should have a comparison/filter
        assert "=" in sql or "WHERE" in sql_upper



"""Test multi-edge-type support for relationship queries.

This test file validates that the transpiler correctly handles queries
with multiple relationship types using the pipe syntax: [:TYPE1|TYPE2].

Current status:
- Variable-length paths with multi-edge types: SUPPORTED
- Fixed-length relationships with multi-edge types: NOT YET SUPPORTED
"""

from gsql2rsql import OpenCypherParser, LogicalPlan, SQLRenderer
from gsql2rsql.common.schema import (
    SimpleGraphSchemaProvider,
    NodeSchema,
    EdgeSchema,
    EntityProperty,
)
from gsql2rsql.renderer.schema_provider import (
    SimpleSQLSchemaProvider,
    SQLTableDescriptor,
)

from tests.utils.sql_assertions import (
    assert_has_select,
    assert_has_recursive_cte,
    assert_cycle_detection,
    SQLStructure,
)


class TestMultiEdgeTypeSupport:
    """Test multi-edge-type support for various query patterns."""

    def setup_method(self) -> None:
        """Set up test fixtures with single edge table schema."""
        # Graph schema with two edge types between Person nodes
        self.graph_schema = SimpleGraphSchemaProvider()
        self.graph_schema.add_node(
            NodeSchema(
                name="Person",
                properties=[
                    EntityProperty("id", int),
                    EntityProperty("name", str),
                ],
                node_id_property=EntityProperty("id", int),
            )
        )
        self.graph_schema.add_edge(
            EdgeSchema(
                name="KNOWS",
                source_node_id="Person",
                sink_node_id="Person",
                source_id_property=EntityProperty("source_id", int),
                sink_id_property=EntityProperty("target_id", int),
            )
        )
        self.graph_schema.add_edge(
            EdgeSchema(
                name="FOLLOWS",
                source_node_id="Person",
                sink_node_id="Person",
                source_id_property=EntityProperty("source_id", int),
                sink_id_property=EntityProperty("target_id", int),
            )
        )

        # SQL schema with single edge table and type filters
        self.sql_schema = SimpleSQLSchemaProvider()
        self.sql_schema.add_node(
            NodeSchema(
                name="Person",
                properties=[
                    EntityProperty("id", int),
                    EntityProperty("name", str),
                ],
                node_id_property=EntityProperty("id", int),
            ),
            SQLTableDescriptor(table_name="graph.Person"),
        )
        self.sql_schema.add_edge(
            EdgeSchema(
                name="KNOWS",
                source_node_id="Person",
                sink_node_id="Person",
                source_id_property=EntityProperty("source_id", int),
                sink_id_property=EntityProperty("target_id", int),
            ),
            SQLTableDescriptor(
                table_name="graph.edges",
                filter="edge_type = 'KNOWS'",
            ),
        )
        self.sql_schema.add_edge(
            EdgeSchema(
                name="FOLLOWS",
                source_node_id="Person",
                sink_node_id="Person",
                source_id_property=EntityProperty("source_id", int),
                sink_id_property=EntityProperty("target_id", int),
            ),
            SQLTableDescriptor(
                table_name="graph.edges",
                filter="edge_type = 'FOLLOWS'",
            ),
        )

    def _transpile(self, cypher: str) -> str:
        """Helper to transpile a Cypher query."""
        parser = OpenCypherParser()
        ast = parser.parse(cypher)
        plan = LogicalPlan.process_query_tree(ast, self.graph_schema)
        plan.resolve(original_query=cypher)
        renderer = SQLRenderer(db_schema_provider=self.sql_schema)
        return renderer.render_plan(plan)

    # =========================================================================
    # Variable-Length Multi-Edge Type Tests (SUPPORTED)
    # =========================================================================

    def test_variable_length_multi_edge_generates_recursive_cte(self) -> None:
        """Test that [:KNOWS|FOLLOWS*1..3] generates WITH RECURSIVE."""
        cypher = """
        MATCH (p:Person)-[:KNOWS|FOLLOWS*1..3]->(f:Person)
        RETURN DISTINCT f.name
        """
        sql = self._transpile(cypher)

        assert_has_recursive_cte(sql)
        assert_has_select(sql)

    def test_variable_length_multi_edge_includes_both_types(self) -> None:
        """Test that SQL references both edge types."""
        cypher = """
        MATCH (p:Person)-[:KNOWS|FOLLOWS*1..3]->(f:Person)
        RETURN f.id
        """
        sql = self._transpile(cypher)

        # Should contain references to both edge types
        assert "KNOWS" in sql, "Expected KNOWS edge type in SQL"
        assert "FOLLOWS" in sql, "Expected FOLLOWS edge type in SQL"

    def test_variable_length_multi_edge_uses_or_filter(self) -> None:
        """Test that multi-edge uses combined OR filter for single table."""
        cypher = """
        MATCH (p:Person)-[:KNOWS|FOLLOWS*1..3]->(f:Person)
        RETURN f.name
        """
        sql = self._transpile(cypher)

        # When both types are in same table, should use OR
        # Either explicit OR or IN clause is acceptable
        sql_upper = sql.upper()
        has_or = " OR " in sql_upper
        has_in = "IN (" in sql_upper

        assert has_or or has_in, (
            "Expected OR filter or IN clause for combined edge types"
        )

    def test_variable_length_multi_edge_has_cycle_detection(self) -> None:
        """Test that recursive query includes cycle detection."""
        cypher = """
        MATCH (p:Person)-[:KNOWS|FOLLOWS*1..5]->(f:Person)
        RETURN f.id
        """
        sql = self._transpile(cypher)

        assert_cycle_detection(sql)

    def test_variable_length_multi_edge_respects_depth(self) -> None:
        """Test that recursive query respects max depth."""
        cypher = """
        MATCH (p:Person)-[:KNOWS|FOLLOWS*1..4]->(f:Person)
        RETURN f.id
        """
        sql = self._transpile(cypher)

        # Should have depth < 4 check
        assert "depth < 4" in sql.lower(), "Expected depth < 4 limit"

    def test_variable_length_single_edge_type_still_works(self) -> None:
        """Test that single edge type in variable-length still works."""
        cypher = """
        MATCH (p:Person)-[:KNOWS*1..3]->(f:Person)
        RETURN f.name
        """
        sql = self._transpile(cypher)

        assert_has_recursive_cte(sql)
        assert "KNOWS" in sql
        # Should NOT have FOLLOWS when only KNOWS is specified
        assert "FOLLOWS" not in sql

    # =========================================================================
    # Edge Case Tests
    # =========================================================================

    def test_multi_edge_with_where_clause(self) -> None:
        """Test multi-edge traversal with WHERE filter."""
        cypher = """
        MATCH (p:Person)-[:KNOWS|FOLLOWS*1..3]->(f:Person)
        WHERE p.id = 1
        RETURN f.name
        """
        sql = self._transpile(cypher)

        assert_has_recursive_cte(sql)
        # Should have WHERE somewhere in the query
        structure = SQLStructure(raw_sql=sql)
        # Note: WHERE might be in CTE or outer query
        assert "WHERE" in sql.upper()

    def test_multi_edge_with_distinct(self) -> None:
        """Test multi-edge traversal with DISTINCT."""
        cypher = """
        MATCH (p:Person)-[:KNOWS|FOLLOWS*1..3]->(f:Person)
        RETURN DISTINCT f.id
        """
        sql = self._transpile(cypher)

        assert_has_recursive_cte(sql)
        assert "DISTINCT" in sql.upper()



"""Test: BFS (Breadth-First Search) via OpenCypher VLP em PySpark.

Cria um grafo em árvore e verifica que VLP retorna os caminhos corretos.

Grafo de teste (árvore):

       A (root)
      / \
     B   C
    /|   |\
   D E   F G

Arestas (direcionadas):
  A -> B, A -> C
  B -> D, B -> E
  C -> F, C -> G
"""

import pytest
from pyspark.sql import SparkSession
from gsql2rsql import GraphContext


@pytest.fixture(scope="module")
def spark():
    """Create SparkSession for tests."""
    spark = (
        SparkSession.builder
        .appName("BFS_Test")
        .master("local[2]")
        .config("spark.sql.shuffle.partitions", "2")
        .config("spark.driver.memory", "1g")
        .getOrCreate()
    )
    yield spark
    spark.stop()


@pytest.fixture(scope="module")
def graph_data(spark):
    """Create tree graph data."""
    # Vertices: A, B, C, D, E, F, G
    nodes_data = [
        ("A", "Root", "root"),
        ("B", "Child", "level1"),
        ("C", "Child", "level1"),
        ("D", "Leaf", "level2"),
        ("E", "Leaf", "level2"),
        ("F", "Leaf", "level2"),
        ("G", "Leaf", "level2"),
    ]
    nodes_df = spark.createDataFrame(
        nodes_data, ["node_id", "node_type", "level"]
    )
    nodes_df.createOrReplaceTempView("nodes")

    # Edges: parent -> child
    edges_data = [
        ("A", "B", "PARENT_OF"),
        ("A", "C", "PARENT_OF"),
        ("B", "D", "PARENT_OF"),
        ("B", "E", "PARENT_OF"),
        ("C", "F", "PARENT_OF"),
        ("C", "G", "PARENT_OF"),
    ]
    edges_df = spark.createDataFrame(
        edges_data, ["src", "dst", "relationship_type"]
    )
    edges_df.createOrReplaceTempView("edges")

    return {"nodes": nodes_df, "edges": edges_df}


@pytest.fixture(scope="module")
def graph_context(spark, graph_data):
    """Create GraphContext for the tree graph."""
    graph = GraphContext(
        spark=spark,
        nodes_table="nodes",
        edges_table="edges",
        node_type_col="node_type",
        node_id_col="node_id",
        extra_node_attrs={"level": str},
    )
    graph.set_types(
        node_types=["Root", "Child", "Leaf"],
        edge_types=["PARENT_OF"],
    )
    return graph


class TestBFSViaCypher:
    """Testes de BFS usando OpenCypher VLP."""

    def test_bfs_depth_1_from_root(self, spark, graph_context):
        """BFS depth 1: A -> {B, C}."""
        query = """
        MATCH (root)-[:PARENT_OF*1..1]->(child)
        WHERE root.node_id = 'A'
        RETURN root.node_id AS start, child.node_id AS end_node
        ORDER BY end_node
        """

        sql = graph_context.transpile(query)
        print("\n=== SQL para BFS depth 1 ===")
        print(sql)

        result = spark.sql(sql)
        rows = result.collect()

        assert result is not None, "DataFrame não deve ser nulo"
        assert len(rows) == 2, f"Esperado 2 filhos de A, encontrado {len(rows)}"

        children = {row["end_node"] for row in rows}
        assert children == {"B", "C"}, f"Filhos de A devem ser B e C, encontrado {children}"

        print(f"✅ BFS depth 1: A -> {children}")

    def test_bfs_depth_2_from_root(self, spark, graph_context):
        """BFS depth 2: A -> {B, C, D, E, F, G}."""
        query = """
        MATCH path = (root)-[:PARENT_OF*1..2]->(descendant)
        WHERE root.node_id = 'A'
        RETURN root.node_id AS start, descendant.node_id AS end_node, length(path) AS depth
        ORDER BY depth, end_node
        """

        sql = graph_context.transpile(query)
        print("\n=== SQL para BFS depth 2 ===")
        print(sql)

        result = spark.sql(sql)
        rows = result.collect()

        assert result is not None, "DataFrame não deve ser nulo"
        assert len(rows) == 6, f"Esperado 6 descendentes de A, encontrado {len(rows)}"

        # Verificar por depth
        depth_1 = {row["end_node"] for row in rows if row["depth"] == 1}
        depth_2 = {row["end_node"] for row in rows if row["depth"] == 2}

        assert depth_1 == {"B", "C"}, f"Depth 1 deve ser {{B, C}}, encontrado {depth_1}"
        assert depth_2 == {"D", "E", "F", "G"}, f"Depth 2 deve ser {{D, E, F, G}}, encontrado {depth_2}"

        print(f"✅ BFS depth 1: {depth_1}")
        print(f"✅ BFS depth 2: {depth_2}")

    def test_bfs_specific_paths(self, spark, graph_context):
        """Verificar caminhos específicos: A->B->D, A->B->E, A->C->F, A->C->G."""
        query = """
        MATCH path = (root)-[:PARENT_OF*2..2]->(leaf)
        WHERE root.node_id = 'A'
        RETURN nodes(path) AS path_nodes
        """

        sql = graph_context.transpile(query)
        print("\n=== SQL para paths específicos ===")
        print(sql)

        result = spark.sql(sql)
        rows = result.collect()

        assert result is not None, "DataFrame não deve ser nulo"
        assert len(rows) == 4, f"Esperado 4 caminhos de tamanho 2, encontrado {len(rows)}"

        # Extrair caminhos
        paths = [tuple(row["path_nodes"]) for row in rows]
        expected_paths = [
            ("A", "B", "D"),
            ("A", "B", "E"),
            ("A", "C", "F"),
            ("A", "C", "G"),
        ]

        for expected in expected_paths:
            assert expected in paths, f"Caminho {expected} não encontrado em {paths}"

        print(f"✅ Caminhos encontrados: {paths}")

    def test_bfs_from_intermediate_node(self, spark, graph_context):
        """BFS a partir de nó intermediário: B -> {D, E}."""
        query = """
        MATCH (start)-[:PARENT_OF*1..1]->(child)
        WHERE start.node_id = 'B'
        RETURN start.node_id AS start, child.node_id AS end_node
        ORDER BY end_node
        """

        sql = graph_context.transpile(query)
        print("\n=== SQL para BFS de B ===")
        print(sql)

        result = spark.sql(sql)
        rows = result.collect()

        assert result is not None
        assert len(rows) == 2, f"B tem 2 filhos, encontrado {len(rows)}"

        children = {row["end_node"] for row in rows}
        assert children == {"D", "E"}, f"Filhos de B devem ser D e E, encontrado {children}"

        print(f"✅ BFS de B: {children}")

    def test_bfs_leaf_node_no_children(self, spark, graph_context):
        """BFS de nó folha deve retornar vazio."""
        query = """
        MATCH (leaf)-[:PARENT_OF*1..1]->(child)
        WHERE leaf.node_id = 'D'
        RETURN leaf.node_id AS start, child.node_id AS end_node
        """

        sql = graph_context.transpile(query)
        print("\n=== SQL para BFS de folha D ===")
        print(sql)

        result = spark.sql(sql)
        rows = result.collect()

        assert result is not None
        assert len(rows) == 0, f"D é folha, não deve ter filhos. Encontrado {len(rows)}"

        print("✅ BFS de folha D: nenhum filho (correto)")

    def test_bfs_unlabeled_nodes(self, spark, graph_context):
        """BFS com nós sem label (wildcard)."""
        query = """
        MATCH (start)-[:PARENT_OF*1..2]->(end_node)
        WHERE start.node_id = 'A'
        RETURN start.node_id AS root, end_node.node_id AS descendant, end_node.level AS level
        ORDER BY level, descendant
        """

        sql = graph_context.transpile(query)
        print("\n=== SQL para BFS sem labels ===")
        print(sql)

        result = spark.sql(sql)
        rows = result.collect()

        assert result is not None
        assert len(rows) == 6, f"Esperado 6 descendentes, encontrado {len(rows)}"

        # Verificar levels
        level1_nodes = {row["descendant"] for row in rows if row["level"] == "level1"}
        level2_nodes = {row["descendant"] for row in rows if row["level"] == "level2"}

        assert level1_nodes == {"B", "C"}
        assert level2_nodes == {"D", "E", "F", "G"}

        print(f"✅ Level 1: {level1_nodes}")
        print(f"✅ Level 2: {level2_nodes}")



"""Test 30: Math functions.

Validates mathematical functions transpilation from Cypher to Databricks SQL.

Cypher math functions map directly to Databricks SQL equivalents:
- abs(x) -> ABS(x)
- ceil(x) -> CEIL(x)
- floor(x) -> FLOOR(x)
- round(x) -> ROUND(x)
- round(x, n) -> ROUND(x, n)
- sqrt(x) -> SQRT(x)
- sign(x) -> SIGN(x)
- log(x) -> LN(x)  # Natural log
- log10(x) -> LOG10(x)
- exp(x) -> EXP(x)
- sin(x) -> SIN(x)
- cos(x) -> COS(x)
- tan(x) -> TAN(x)
- rand() -> RAND()
- pi() -> PI()
- e() -> E()

Fraud Use Cases:
- Calculate absolute differences between expected and actual amounts
- Round transaction amounts for reporting
- Compute statistical measures
"""

from gsql2rsql import OpenCypherParser, LogicalPlan, SQLRenderer
from gsql2rsql.common.schema import (
    SimpleGraphSchemaProvider,
    NodeSchema,
    EntityProperty,
)
from gsql2rsql.renderer.schema_provider import (
    SimpleSQLSchemaProvider,
    SQLTableDescriptor,
)


class TestMathFunctions:
    """Test math function transpilation."""

    TEST_ID = "30"
    TEST_NAME = "math_functions"

    def setup_method(self) -> None:
        """Set up test fixtures."""
        self.graph_schema = SimpleGraphSchemaProvider()
        self.graph_schema.add_node(
            NodeSchema(
                name="Account",
                properties=[
                    EntityProperty("id", str),
                    EntityProperty("balance", float),
                    EntityProperty("score", float),
                ],
                node_id_property=EntityProperty("id", str),
            )
        )

        self.sql_schema = SimpleSQLSchemaProvider()
        self.sql_schema.add_node(
            NodeSchema(
                name="Account",
                properties=[
                    EntityProperty("id", str),
                    EntityProperty("balance", float),
                    EntityProperty("score", float),
                ],
                node_id_property=EntityProperty("id", str),
            ),
            SQLTableDescriptor(table_name="graph.Account"),
        )

    def _transpile(self, cypher: str) -> str:
        """Helper to transpile a Cypher query."""
        parser = OpenCypherParser()
        ast = parser.parse(cypher)
        plan = LogicalPlan.process_query_tree(ast, self.graph_schema)
        plan.resolve(original_query=cypher)
        renderer = SQLRenderer(db_schema_provider=self.sql_schema)
        return renderer.render_plan(plan)

    def test_abs_function(self) -> None:
        """Test ABS function."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, abs(a.balance) AS absBalance
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "ABS(" in sql_upper, "Should use ABS function"

    def test_ceil_function(self) -> None:
        """Test CEIL function."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, ceil(a.balance) AS ceilBalance
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "CEIL(" in sql_upper, "Should use CEIL function"

    def test_floor_function(self) -> None:
        """Test FLOOR function."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, floor(a.balance) AS floorBalance
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "FLOOR(" in sql_upper, "Should use FLOOR function"

    def test_round_function(self) -> None:
        """Test ROUND function."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, round(a.balance) AS roundBalance
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "ROUND(" in sql_upper, "Should use ROUND function"

    def test_round_with_precision(self) -> None:
        """Test ROUND function with precision."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, round(a.balance, 2) AS roundBalance
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "ROUND(" in sql_upper, "Should use ROUND function"
        assert "2" in sql, "Should have precision parameter"

    def test_sqrt_function(self) -> None:
        """Test SQRT function."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, sqrt(a.score) AS sqrtScore
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "SQRT(" in sql_upper, "Should use SQRT function"

    def test_sign_function(self) -> None:
        """Test SIGN function."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, sign(a.balance) AS signBalance
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "SIGN(" in sql_upper, "Should use SIGN function"

    def test_log_function(self) -> None:
        """Test LOG (natural log) function.

        Cypher log() maps to Databricks LN().
        """
        cypher = """
        MATCH (a:Account)
        RETURN a.id, log(a.score) AS logScore
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "LN(" in sql_upper, "Should use LN function for natural log"

    def test_log10_function(self) -> None:
        """Test LOG10 function."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, log10(a.score) AS log10Score
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "LOG10(" in sql_upper, "Should use LOG10 function"

    def test_exp_function(self) -> None:
        """Test EXP function."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, exp(a.score) AS expScore
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "EXP(" in sql_upper, "Should use EXP function"

    def test_sin_function(self) -> None:
        """Test SIN function."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, sin(a.score) AS sinScore
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "SIN(" in sql_upper, "Should use SIN function"

    def test_cos_function(self) -> None:
        """Test COS function."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, cos(a.score) AS cosScore
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "COS(" in sql_upper, "Should use COS function"

    def test_tan_function(self) -> None:
        """Test TAN function."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, tan(a.score) AS tanScore
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "TAN(" in sql_upper, "Should use TAN function"

    def test_rand_function(self) -> None:
        """Test RAND function."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, rand() AS randomValue
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "RAND()" in sql_upper, "Should use RAND function"

    def test_pi_function(self) -> None:
        """Test PI function."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, pi() AS piValue
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "PI()" in sql_upper, "Should use PI function"

    def test_e_function(self) -> None:
        """Test E function (Euler's number)."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, e() AS eulerValue
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "E()" in sql_upper, "Should use E function"

    def test_combined_math_expression(self) -> None:
        """Test combined math expression."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, abs(a.balance) + sqrt(a.score) AS combined
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "ABS(" in sql_upper
        assert "SQRT(" in sql_upper

    def test_nested_math_functions(self) -> None:
        """Test nested math functions."""
        cypher = """
        MATCH (a:Account)
        RETURN a.id, round(sqrt(abs(a.balance)), 2) AS nested
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "ROUND(" in sql_upper
        assert "SQRT(" in sql_upper
        assert "ABS(" in sql_upper

    def test_fraud_abs_difference(self) -> None:
        """Test fraud scenario: calculate absolute difference."""
        cypher = """
        MATCH (a:Account)
        WHERE abs(a.balance - 10000) < 100
        RETURN a.id
        """
        sql = self._transpile(cypher)

        sql_upper = sql.upper()
        assert "ABS(" in sql_upper



"""ANTLR4 visitor for converting parse tree to AST."""

from __future__ import annotations

from typing import Any

from gsql2rsql.common.exceptions import TranspilerSyntaxErrorException
from gsql2rsql.common.logging import ILoggable
from gsql2rsql.parser.ast import (
    Entity,
    InfixOperator,
    InfixQueryNode,
    LimitClause,
    ListPredicateType,
    MatchClause,
    NodeEntity,
    PartialQueryNode,
    QueryExpression,
    QueryExpressionAggregationFunction,
    QueryExpressionBinary,
    QueryExpressionCaseExpression,
    QueryExpressionExists,
    QueryExpressionFunction,
    QueryExpressionList,
    QueryExpressionListComprehension,
    QueryExpressionListPredicate,
    QueryExpressionMapLiteral,
    QueryExpressionParameter,
    QueryExpressionProperty,
    QueryExpressionReduce,
    QueryExpressionValue,
    QueryExpressionWithAlias,
    QueryNode,
    RelationshipDirection,
    RelationshipEntity,
    SingleQueryNode,
    SortItem,
    SortOrder,
    UnwindClause,
)
from gsql2rsql.parser.operators import (
    AggregationFunction,
    Function,
    try_get_function,
    try_get_operator,
    try_parse_aggregation_function,
)


class CypherVisitor:
    """
    Visitor that converts ANTLR parse tree to openCypher AST.

    This class implements the visitor pattern to walk the ANTLR-generated
    parse tree and construct our custom AST representation.
    """

    def __init__(self, logger: ILoggable | None = None) -> None:
        self._logger = logger

    def visit(self, tree: Any) -> Any:
        """
        Visit a parse tree node and dispatch to the appropriate handler.

        This is the main entry point for visiting parse tree nodes.
        """
        if tree is None:
            return None

        # Get the visitor method based on the node type
        class_name = tree.__class__.__name__

        # Handle context names like 'OC_CypherContext' -> 'visit_oC_Cypher'
        if class_name.endswith("Context"):
            # Remove 'Context' suffix and convert 'OC_' to 'oC_'
            rule_name = class_name[:-7]
            if rule_name.startswith("OC_"):
                rule_name = "oC_" + rule_name[3:]
            method_name = f"visit_{rule_name}"
        else:
            method_name = f"visit_{class_name}"

        visitor_method = getattr(self, method_name, None)

        if visitor_method:
            return visitor_method(tree)

        # Default: visit children and return last result
        return self.visit_children(tree)

    def visit_children(self, tree: Any) -> Any:
        """Visit all children of a node and return the last result."""
        result: Any = None
        for i in range(tree.getChildCount()):
            child = tree.getChild(i)
            # Skip whitespace/terminal nodes
            if not hasattr(child, "getRuleIndex"):
                continue
            child_result = self.visit(child)
            if child_result is not None:
                result = child_result
        return result

    # =========================================================================
    # Top-level query visitors
    # =========================================================================

    def visit_oC_Cypher(self, ctx: Any) -> QueryNode:
        """Visit the root Cypher context."""
        return self.visit(ctx.oC_Statement())

    def visit_oC_Statement(self, ctx: Any) -> QueryNode:
        """Visit a statement context."""
        return self.visit(ctx.oC_Query())

    def visit_oC_Query(self, ctx: Any) -> QueryNode:
        """Visit a query context."""
        if ctx.oC_RegularQuery():
            return self.visit(ctx.oC_RegularQuery())
        return SingleQueryNode()

    def visit_oC_RegularQuery(self, ctx: Any) -> QueryNode:
        """Visit a regular query context."""
        result: QueryNode = self.visit(ctx.oC_SingleQuery())

        # Handle UNION clauses
        for union_ctx in ctx.oC_Union() or []:
            right = self.visit(union_ctx.oC_SingleQuery())
            has_all = any(
                str(child.getText()).upper() == "ALL"
                for i in range(union_ctx.getChildCount())
                if hasattr((child := union_ctx.getChild(i)), "getText")
            )
            operator = InfixOperator.UNION_ALL if has_all else InfixOperator.UNION
            result = InfixQueryNode(
                operator=operator,
                left_query=result,
                right_query=right,
            )

        return result

    def visit_oC_SingleQuery(self, ctx: Any) -> SingleQueryNode:
        """Visit a single query context."""
        parts: list[PartialQueryNode] = []

        # Visit single part query if present
        if ctx.oC_SinglePartQuery():
            part = self.visit(ctx.oC_SinglePartQuery())
            if isinstance(part, PartialQueryNode):
                parts.append(part)

        # Visit multi-part query if present
        if ctx.oC_MultiPartQuery():
            multi_parts = self.visit(ctx.oC_MultiPartQuery())
            if isinstance(multi_parts, list):
                parts.extend(multi_parts)
            elif isinstance(multi_parts, PartialQueryNode):
                parts.append(multi_parts)

        return SingleQueryNode(parts=parts)

    def visit_oC_MultiPartQuery(self, ctx: Any) -> list[PartialQueryNode]:
        """Visit a multi-part query context.

        Grammar:
        oC_MultiPartQuery
            : ( ( oC_ReadingClause SP? )* ( oC_UpdatingClause SP? )* oC_With SP? )+
              oC_SinglePartQuery ;

        Multi-part queries have WITH clauses that produce intermediate results.
        Each group of reading clauses followed by a WITH forms one PartialQueryNode.

        IMPORTANT: We must iterate through children in order to correctly group
        reading clauses with their corresponding WITH clause. The grammar structure
        means that reading clauses belong to the NEXT WITH clause, not the previous one.
        """
        parts: list[PartialQueryNode] = []

        # Collect reading clauses until we hit a WITH, then create a part
        current_reading_clauses: list[Any] = []

        # Iterate through children in order to properly group
        # reading clauses with their corresponding WITH
        for child in ctx.children or []:
            child_class = child.__class__.__name__

            if child_class == "OC_ReadingClauseContext":
                # Accumulate reading clauses until we hit a WITH
                current_reading_clauses.append(child)

            elif child_class == "OC_WithContext":
                # Create a PartialQueryNode with the accumulated reading clauses
                part = PartialQueryNode()

                # Add all accumulated reading clauses to this part
                for reading_ctx in current_reading_clauses:
                    reading = self.visit(reading_ctx)
                    if isinstance(reading, MatchClause):
                        part.match_clauses.append(reading)
                    elif isinstance(reading, UnwindClause):
                        part.unwind_clauses.append(reading)

                # Process the WITH clause
                with_result = self.visit(child)
                if isinstance(with_result, dict):
                    self._apply_return_result(part, with_result)

                parts.append(part)

                # Clear reading clauses - they've been consumed by this WITH
                current_reading_clauses = []

            elif child_class == "OC_SinglePartQueryContext":
                # Final single part query - should have no reading clauses before it
                # (they should have been consumed by WITH clauses)
                final_part = self.visit(child)
                if isinstance(final_part, PartialQueryNode):
                    # If there are leftover reading clauses, add them to the final part
                    # This shouldn't normally happen with well-formed queries
                    for reading_ctx in current_reading_clauses:
                        reading = self.visit(reading_ctx)
                        if isinstance(reading, MatchClause):
                            final_part.match_clauses.insert(0, reading)
                        elif isinstance(reading, UnwindClause):
                            final_part.unwind_clauses.insert(0, reading)
                    parts.append(final_part)
                current_reading_clauses = []

            # Skip other children like whitespace (SP tokens)

        return parts

    def visit_oC_With(self, ctx: Any) -> dict[str, Any]:
        """Visit a WITH clause context.

        Grammar:
        oC_With: WITH oC_ProjectionBody ( SP? oC_Where )? ;

        WITH is like RETURN but passes results to the next part of the query.
        The optional WHERE clause acts like HAVING when filtering aggregated columns.
        """
        result: dict[str, Any] = {}

        # Process the projection body (same as RETURN)
        if ctx.oC_ProjectionBody():
            body_result = self.visit(ctx.oC_ProjectionBody())
            if isinstance(body_result, dict):
                result.update(body_result)

        # Process the optional WHERE clause (acts as HAVING for aggregations)
        if ctx.oC_Where():
            where_expr = self.visit(ctx.oC_Where())
            if isinstance(where_expr, QueryExpression):
                # Store the WHERE/HAVING expression
                # This will be used to generate HAVING clause in SQL
                result["having"] = where_expr

        return result

    def visit_oC_SinglePartQuery(self, ctx: Any) -> PartialQueryNode:
        """Visit a single part query context."""
        part = PartialQueryNode()

        # Visit reading clauses (MATCH and UNWIND)
        for reading_ctx in ctx.oC_ReadingClause() or []:
            reading = self.visit(reading_ctx)
            if isinstance(reading, MatchClause):
                part.match_clauses.append(reading)
            elif isinstance(reading, UnwindClause):
                part.unwind_clauses.append(reading)

        # Visit return clause
        if ctx.oC_Return():
            return_result = self.visit(ctx.oC_Return())
            if return_result:
                self._apply_return_result(part, return_result)

        return part

    def _apply_return_result(self, part: PartialQueryNode, result: dict[str, Any]) -> None:
        """Apply return clause results to a partial query node."""
        if "body" in result:
            part.return_body = result["body"]
        if "distinct" in result:
            part.is_distinct = result["distinct"]
        if "order" in result:
            part.order_by = result["order"]
        if "limit" in result:
            part.limit_clause = result["limit"]
        if "having" in result:
            part.having_expression = result["having"]

    # =========================================================================
    # Reading clause visitors
    # =========================================================================

    def visit_oC_ReadingClause(self, ctx: Any) -> MatchClause | UnwindClause | None:
        """Visit a reading clause context.

        Grammar:
        oC_ReadingClause: oC_Match | oC_Unwind | oC_InQueryCall ;
        """
        if ctx.oC_Match():
            return self.visit(ctx.oC_Match())
        if ctx.oC_Unwind():
            return self.visit(ctx.oC_Unwind())
        return None

    def visit_oC_Unwind(self, ctx: Any) -> UnwindClause:
        """Visit an UNWIND clause context.

        Grammar:
        oC_Unwind: UNWIND SP oC_Expression SP AS SP oC_Variable ;

        Example: UNWIND a.tags AS tag
        """
        # Get the expression to unwind
        list_expression = self.visit(ctx.oC_Expression())

        # Get the variable name
        variable_name = ctx.oC_Variable().getText()

        return UnwindClause(
            list_expression=list_expression,
            variable_name=variable_name,
        )

    def visit_oC_Match(self, ctx: Any) -> MatchClause:
        """Visit a MATCH clause context."""
        is_optional = any(
            str(child.getText()).upper() == "OPTIONAL"
            for i in range(ctx.getChildCount())
            if hasattr((child := ctx.getChild(i)), "getText")
            and not hasattr(child, "getRuleIndex")
        )

        # Extract pattern and path variable
        entities: list[Entity] = []
        path_variable = ""

        if ctx.oC_Pattern():
            for part_ctx in ctx.oC_Pattern().oC_PatternPart() or []:
                # Check for path variable assignment: path = (a)-[]->(b)
                if part_ctx.oC_Variable():
                    path_variable = part_ctx.oC_Variable().getText()

                # Get the pattern elements
                if part_ctx.oC_AnonymousPatternPart():
                    part = self.visit(part_ctx.oC_AnonymousPatternPart())
                    if isinstance(part, list):
                        entities.extend(part)
                    elif isinstance(part, Entity):
                        entities.append(part)

        where_expr = self.visit(ctx.oC_Where()) if ctx.oC_Where() else None

        return MatchClause(
            pattern_parts=entities,
            is_optional=is_optional,
            where_expression=where_expr,
            path_variable=path_variable,
        )

    def visit_oC_Pattern(self, ctx: Any) -> list[Entity]:
        """Visit a pattern context."""
        entities: list[Entity] = []
        for part_ctx in ctx.oC_PatternPart() or []:
            part = self.visit(part_ctx)
            if isinstance(part, list):
                entities.extend(part)
            elif isinstance(part, Entity):
                entities.append(part)
        return entities

    def visit_oC_PatternPart(self, ctx: Any) -> list[Entity]:
        """Visit a pattern part context.

        Note: Path variable extraction is handled in visit_oC_Match for MatchClause.
        This method is used for other contexts that don't need path variables.
        """
        if ctx.oC_AnonymousPatternPart():
            return self.visit(ctx.oC_AnonymousPatternPart())
        return []

    def visit_oC_AnonymousPatternPart(self, ctx: Any) -> list[Entity]:
        """Visit an anonymous pattern part context."""
        return self.visit(ctx.oC_PatternElement())

    def visit_oC_PatternElement(self, ctx: Any) -> list[Entity]:
        """Visit a pattern element context."""
        entities: list[Entity] = []

        # Get the node pattern
        if ctx.oC_NodePattern():
            node = self.visit(ctx.oC_NodePattern())
            if isinstance(node, NodeEntity):
                entities.append(node)

        # Get the chain elements
        for chain_ctx in ctx.oC_PatternElementChain() or []:
            chain = self.visit(chain_ctx)
            if isinstance(chain, list):
                entities.extend(chain)

        # Handle nested pattern element
        if ctx.oC_PatternElement():
            nested = self.visit(ctx.oC_PatternElement())
            if isinstance(nested, list):
                entities.extend(nested)

        return entities

    def visit_oC_NodePattern(self, ctx: Any) -> NodeEntity:
        """Visit a node pattern context.

        Extracts node alias, labels, and inline property filters.

        Grammar:
            oC_NodePattern: '(' SP? ( oC_Variable SP? )? ( oC_NodeLabels SP? )?
                           ( oC_Properties SP? )? ')' ;

        Example patterns:
            (n)                          # Just variable
            (n:Person)                   # Variable + label
            (n:Person {name: 'Alice'})   # Variable + label + properties
            (:Person {age: 30})          # Label + properties, no variable
        """
        alias = ""
        entity_name = ""
        inline_properties = None

        if ctx.oC_Variable():
            alias = ctx.oC_Variable().getText()

        if ctx.oC_NodeLabels():
            labels = ctx.oC_NodeLabels().getText()
            # Remove leading colon
            entity_name = labels.lstrip(":")
            # Take first label if multiple
            if ":" in entity_name:
                entity_name = entity_name.split(":")[0]

        # NEW: Extract inline properties if present
        if ctx.oC_Properties():
            # Visit oC_Properties (which can be oC_MapLiteral or oC_Parameter)
            props = self.visit(ctx.oC_Properties())
            # Only store if it's a map literal (not a parameter)
            if isinstance(props, QueryExpressionMapLiteral):
                inline_properties = props

        return NodeEntity(
            alias=alias,
            entity_name=entity_name,
            inline_properties=inline_properties
        )

    def visit_oC_PatternElementChain(self, ctx: Any) -> list[Entity]:
        """Visit a pattern element chain context."""
        entities: list[Entity] = []

        if ctx.oC_RelationshipPattern():
            rel = self.visit(ctx.oC_RelationshipPattern())
            if isinstance(rel, RelationshipEntity):
                entities.append(rel)

        if ctx.oC_NodePattern():
            node = self.visit(ctx.oC_NodePattern())
            if isinstance(node, NodeEntity):
                # Update relationship's connected entities
                if entities and isinstance(entities[-1], RelationshipEntity):
                    entities[-1].right_entity_name = node.entity_name
                entities.append(node)

        return entities

    def visit_oC_RelationshipPattern(self, ctx: Any) -> RelationshipEntity:
        """Visit a relationship pattern context.

        Extracts relationship alias, type, direction, variable-length path
        info, and inline property filters.

        Grammar:
            oC_RelationshipPattern: oC_LeftArrowHead? SP? oC_Dash
                                   oC_RelationshipDetail? oC_Dash SP?
                                   oC_RightArrowHead? ;
            oC_RelationshipDetail: '[' SP? oC_Variable? SP?
                                  oC_RelationshipTypes? SP?
                                  oC_RangeLiteral? SP?
                                  oC_Properties? SP? ']' ;

        Example patterns:
            -[r:KNOWS]->                              # Basic relationship
            -[r:KNOWS {since: 2020}]->                # With properties
            -[r:KNOWS*1..3]->                         # Variable-length
            -[r:KNOWS*1..3 {weight: 1.0}]->          # Var-length + props
        """
        # Determine direction
        text = ctx.getText()
        if "->" in text:
            direction = RelationshipDirection.FORWARD
        elif "<-" in text:
            direction = RelationshipDirection.BACKWARD
        else:
            direction = RelationshipDirection.BOTH

        # Get details
        detail_ctx = ctx.oC_RelationshipDetail() if ctx.oC_RelationshipDetail() else None

        alias = ""
        entity_name = ""
        min_hops = None
        max_hops = None
        inline_properties = None

        if detail_ctx:
            if detail_ctx.oC_Variable():
                alias = detail_ctx.oC_Variable().getText()
            if detail_ctx.oC_RelationshipTypes():
                types_text = detail_ctx.oC_RelationshipTypes().getText()
                entity_name = types_text.lstrip(":")

            # Extract variable-length path information (e.g., *1..5)
            if detail_ctx.oC_RangeLiteral():
                range_ctx = detail_ctx.oC_RangeLiteral()
                # Parse the range literal - format: *[min][..[max]]
                integers = range_ctx.oC_IntegerLiteral()
                if integers:
                    if len(integers) >= 1:
                        min_hops = int(integers[0].getText())
                    if len(integers) >= 2:
                        max_hops = int(integers[1].getText())
                    elif ".." in range_ctx.getText():
                        # *1.. means min_hops only, unlimited max
                        pass
                    else:
                        # *3 means exactly 3 hops (min=max=3)
                        max_hops = min_hops
                else:
                    # Just * means any number of hops (default 1..unlimited)
                    min_hops = 1

            # NEW: Extract inline properties if present
            if detail_ctx.oC_Properties():
                props = self.visit(detail_ctx.oC_Properties())
                if isinstance(props, QueryExpressionMapLiteral):
                    inline_properties = props

        return RelationshipEntity(
            alias=alias,
            entity_name=entity_name,
            direction=direction,
            min_hops=min_hops,
            max_hops=max_hops,
            inline_properties=inline_properties,
        )

    # =========================================================================
    # WHERE clause visitor
    # =========================================================================

    def visit_oC_Where(self, ctx: Any) -> QueryExpression | None:
        """Visit a WHERE clause context."""
        if ctx.oC_Expression():
            return self.visit(ctx.oC_Expression())
        return None

    # =========================================================================
    # RETURN clause visitors
    # =========================================================================

    def visit_oC_Return(self, ctx: Any) -> dict[str, Any]:
        """Visit a RETURN clause context."""
        result: dict[str, Any] = {}

        # New grammar uses oC_ProjectionBody directly in oC_Return
        if hasattr(ctx, "oC_ProjectionBody") and ctx.oC_ProjectionBody():
            body_result = self.visit(ctx.oC_ProjectionBody())
            if isinstance(body_result, dict):
                result.update(body_result)
        # Fallback for old grammar structure
        elif hasattr(ctx, "oC_ReturnBody") and ctx.oC_ReturnBody():
            body_result = self.visit(ctx.oC_ReturnBody())
            if isinstance(body_result, dict):
                result.update(body_result)

        return result

    def visit_oC_ProjectionBody(self, ctx: Any) -> dict[str, Any]:
        """Visit a projection body context (new grammar)."""
        result: dict[str, Any] = {}

        # Check for DISTINCT keyword
        text_children = [
            ctx.getChild(i).getText()
            for i in range(ctx.getChildCount())
            if hasattr(ctx.getChild(i), "getText")
            and not hasattr(ctx.getChild(i), "getRuleIndex")
        ]
        result["distinct"] = any(t.upper() == "DISTINCT" for t in text_children)

        if ctx.oC_ProjectionItems():
            items = self.visit(ctx.oC_ProjectionItems())
            if isinstance(items, list):
                result["body"] = items

        if ctx.oC_Order():
            order = self.visit(ctx.oC_Order())
            if isinstance(order, list):
                result["order"] = order

        if ctx.oC_Limit():
            limit = self.visit(ctx.oC_Limit())
            result["limit"] = LimitClause(limit_expression=limit)

        if ctx.oC_Skip():
            skip = self.visit(ctx.oC_Skip())
            if "limit" in result:
                result["limit"].skip_expression = skip
            else:
                result["limit"] = LimitClause(
                    limit_expression=QueryExpressionValue(
                        value=None, value_type=type(None)
                    ),
                    skip_expression=skip,
                )

        return result

    def visit_oC_ReturnBody(self, ctx: Any) -> dict[str, Any]:
        """Visit a return body context (legacy)."""
        result: dict[str, Any] = {}

        if ctx.oC_ReturnItems():
            items = self.visit(ctx.oC_ReturnItems())
            if isinstance(items, list):
                result["body"] = items

        if ctx.oC_Order():
            order = self.visit(ctx.oC_Order())
            if isinstance(order, list):
                result["order"] = order

        if ctx.oC_Limit():
            limit = self.visit(ctx.oC_Limit())
            result["limit"] = LimitClause(limit_expression=limit)

        if ctx.oC_Skip():
            skip = self.visit(ctx.oC_Skip())
            if "limit" in result:
                result["limit"].skip_expression = skip
            else:
                result["limit"] = LimitClause(
                    limit_expression=QueryExpressionValue(
                        value=None, value_type=type(None)
                    ),
                    skip_expression=skip,
                )

        return result

    def visit_oC_ProjectionItems(self, ctx: Any) -> list[QueryExpressionWithAlias]:
        """Visit projection items context."""
        items: list[QueryExpressionWithAlias] = []

        for item_ctx in ctx.oC_ProjectionItem() or []:
            item = self.visit(item_ctx)
            if isinstance(item, QueryExpressionWithAlias):
                items.append(item)

        return items

    def visit_oC_ProjectionItem(self, ctx: Any) -> QueryExpressionWithAlias | None:
        """Visit a projection item context."""
        expr = self.visit(ctx.oC_Expression())

        if not isinstance(expr, QueryExpression):
            return None

        alias = None
        if ctx.oC_Variable():
            alias = ctx.oC_Variable().getText()
        elif isinstance(expr, QueryExpressionProperty):
            alias = expr.property_name or expr.variable_name

        return QueryExpressionWithAlias(inner_expression=expr, alias=alias or "")

    def visit_oC_ReturnItems(self, ctx: Any) -> list[QueryExpressionWithAlias]:
        """Visit return items context (legacy)."""
        items: list[QueryExpressionWithAlias] = []

        for item_ctx in ctx.oC_ReturnItem() or []:
            item = self.visit(item_ctx)
            if isinstance(item, QueryExpressionWithAlias):
                items.append(item)

        return items

    def visit_oC_ReturnItem(self, ctx: Any) -> QueryExpressionWithAlias | None:
        """Visit a return item context (legacy)."""
        expr = self.visit(ctx.oC_Expression())

        if not isinstance(expr, QueryExpression):
            return None

        # Get alias if present
        alias = ""
        if ctx.oC_Variable():
            alias = ctx.oC_Variable().getText()
        elif isinstance(expr, QueryExpressionProperty):
            alias = expr.property_name or expr.variable_name
        else:
            alias = str(expr)

        return QueryExpressionWithAlias(inner_expression=expr, alias=alias)

    def visit_oC_Order(self, ctx: Any) -> list[SortItem]:
        """Visit an ORDER BY clause context."""
        items: list[SortItem] = []

        for sort_ctx in ctx.oC_SortItem() or []:
            item = self.visit(sort_ctx)
            if isinstance(item, SortItem):
                items.append(item)

        return items

    def visit_oC_SortItem(self, ctx: Any) -> SortItem:
        """Visit a sort item context."""
        expr = self.visit(ctx.oC_Expression())

        # Check for DESC
        text_children = [
            ctx.getChild(i).getText()
            for i in range(ctx.getChildCount())
            if hasattr(ctx.getChild(i), "getText")
        ]
        order = (
            SortOrder.DESC
            if any(t.upper() in ("DESC", "DESCENDING") for t in text_children)
            else SortOrder.ASC
        )

        if not isinstance(expr, QueryExpression):
            expr = QueryExpressionValue(value=None, value_type=type(None))

        return SortItem(expression=expr, order=order)

    def visit_oC_Limit(self, ctx: Any) -> QueryExpression:
        """Visit a LIMIT clause context."""
        return self.visit(ctx.oC_Expression())

    def visit_oC_Skip(self, ctx: Any) -> QueryExpression:
        """Visit a SKIP clause context."""
        return self.visit(ctx.oC_Expression())

    # =========================================================================
    # Expression visitors
    # =========================================================================

    def visit_oC_Expression(self, ctx: Any) -> QueryExpression:
        """Visit an expression context."""
        return self.visit(ctx.oC_OrExpression())

    def visit_oC_OrExpression(self, ctx: Any) -> QueryExpression:
        """Visit an OR expression context."""
        return self._visit_binary_expression(ctx, ctx.oC_XorExpression(), "or")

    def visit_oC_XorExpression(self, ctx: Any) -> QueryExpression:
        """Visit an XOR expression context."""
        return self._visit_binary_expression(ctx, ctx.oC_AndExpression(), "xor")

    def visit_oC_AndExpression(self, ctx: Any) -> QueryExpression:
        """Visit an AND expression context."""
        return self._visit_binary_expression(ctx, ctx.oC_NotExpression(), "and")

    def visit_oC_NotExpression(self, ctx: Any) -> QueryExpression:
        """Visit a NOT expression context."""
        # Check if NOT is present
        text = ctx.getText()
        has_not = text.upper().startswith("NOT")

        inner = self.visit(ctx.oC_ComparisonExpression())

        if has_not and isinstance(inner, QueryExpression):
            return QueryExpressionFunction(
                function=Function.NOT,
                parameters=[inner],
                data_type=bool,
            )

        return inner

    def visit_oC_ComparisonExpression(self, ctx: Any) -> QueryExpression:
        """Visit a comparison expression context."""
        # New grammar uses oC_StringListNullPredicateExpression
        if hasattr(ctx, "oC_StringListNullPredicateExpression"):
            left = self.visit(ctx.oC_StringListNullPredicateExpression())
        else:
            # Fallback for old grammar
            left = self.visit(ctx.oC_AddOrSubtractExpression())

        # Handle partial comparison expressions
        if ctx.oC_PartialComparisonExpression():
            for partial_ctx in ctx.oC_PartialComparisonExpression():
                op_text = ""
                for i in range(partial_ctx.getChildCount()):
                    child = partial_ctx.getChild(i)
                    if not hasattr(child, "getRuleIndex"):
                        op_text = child.getText()
                        break

                # Get the right side from partial comparison
                if hasattr(partial_ctx, "oC_StringListNullPredicateExpression"):
                    right = self.visit(
                        partial_ctx.oC_StringListNullPredicateExpression()
                    )
                else:
                    right = self.visit(partial_ctx.oC_AddOrSubtractExpression())

                op_info = try_get_operator(op_text)

                if op_info and isinstance(right, QueryExpression):
                    left = QueryExpressionBinary(
                        operator=op_info,
                        left_expression=(
                            left if isinstance(left, QueryExpression) else None
                        ),
                        right_expression=right,
                    )

        return left

    def visit_oC_StringListNullPredicateExpression(
        self, ctx: Any
    ) -> QueryExpression:
        """Visit a string/list/null predicate expression context."""
        result = self.visit(ctx.oC_AddOrSubtractExpression())

        # Handle IS NULL, IS NOT NULL, IN, etc.
        for i in range(ctx.getChildCount()):
            child = ctx.getChild(i)
            if hasattr(child, "getRuleIndex"):
                child_name = child.__class__.__name__
                if "NullPredicateExpression" in child_name:
                    # IS NULL or IS NOT NULL
                    text = child.getText().upper()
                    if "NOT" in text:
                        result = QueryExpressionFunction(
                            function=Function.IS_NOT_NULL,
                            parameters=(
                                [result]
                                if isinstance(result, QueryExpression)
                                else []
                            ),
                            data_type=bool,
                        )
                    else:
                        result = QueryExpressionFunction(
                            function=Function.IS_NULL,
                            parameters=(
                                [result]
                                if isinstance(result, QueryExpression)
                                else []
                            ),
                            data_type=bool,
                        )
                elif "ListPredicateExpression" in child_name:
                    # IN operator: a.id IN $watchlist or a.id IN [1, 2, 3]
                    # The child contains: SP IN SP? oC_AddOrSubtractExpression
                    right_expr = self.visit(child.oC_AddOrSubtractExpression())
                    if isinstance(right_expr, QueryExpression):
                        op_info = try_get_operator("in")
                        result = QueryExpressionBinary(
                            operator=op_info,
                            left_expression=(
                                result
                                if isinstance(result, QueryExpression)
                                else None
                            ),
                            right_expression=right_expr,
                        )

        return result

    def visit_oC_AddOrSubtractExpression(self, ctx: Any) -> QueryExpression:
        """Visit an add/subtract expression context."""
        return self._visit_chained_binary_expression(
            ctx, ctx.oC_MultiplyDivideModuloExpression(), ["+", "-"]
        )

    def visit_oC_MultiplyDivideModuloExpression(self, ctx: Any) -> QueryExpression:
        """Visit a multiply/divide/modulo expression context."""
        return self._visit_chained_binary_expression(
            ctx, ctx.oC_PowerOfExpression(), ["*", "/", "%"]
        )

    def visit_oC_PowerOfExpression(self, ctx: Any) -> QueryExpression:
        """Visit a power expression context."""
        return self._visit_chained_binary_expression(
            ctx, ctx.oC_UnaryAddOrSubtractExpression(), ["^"]
        )

    def visit_oC_UnaryAddOrSubtractExpression(self, ctx: Any) -> QueryExpression:
        """Visit a unary add/subtract expression context."""
        # New grammar uses oC_NonArithmeticOperatorExpression
        if hasattr(ctx, "oC_NonArithmeticOperatorExpression"):
            inner = self.visit(ctx.oC_NonArithmeticOperatorExpression())
        # Old grammar uses oC_StringListNullOperatorExpression
        elif hasattr(ctx, "oC_StringListNullOperatorExpression"):
            inner = self.visit(ctx.oC_StringListNullOperatorExpression())
        else:
            inner = self.visit_children(ctx)

        # Check for unary operator
        text = ctx.getText()
        if text.startswith("-") and isinstance(inner, QueryExpression):
            return QueryExpressionFunction(
                function=Function.NEGATIVE,
                parameters=[inner],
            )

        return inner

    def visit_oC_NonArithmeticOperatorExpression(
        self, ctx: Any
    ) -> QueryExpression:
        """Visit a non-arithmetic operator expression context (new grammar)."""
        # This goes to oC_Atom directly in new grammar
        result = self.visit(ctx.oC_Atom())

        # Handle property lookups
        if hasattr(ctx, "oC_PropertyLookup") and ctx.oC_PropertyLookup():
            for prop_ctx in ctx.oC_PropertyLookup():
                prop_name = self.visit(prop_ctx)
                if isinstance(result, QueryExpressionProperty) and prop_name:
                    result.property_name = prop_name

        return result

    def visit_oC_PropertyLookup(self, ctx: Any) -> str | None:
        """Visit a property lookup context."""
        if ctx.oC_PropertyKeyName():
            return ctx.oC_PropertyKeyName().getText()
        return None

    def visit_oC_StringListNullOperatorExpression(
        self, ctx: Any
    ) -> QueryExpression:
        """Visit a string/list/null operator expression context (old grammar)."""
        result = self.visit(ctx.oC_PropertyOrLabelsExpression())

        # Handle IS NULL / IS NOT NULL
        for i in range(ctx.getChildCount()):
            child = ctx.getChild(i)
            if hasattr(child, "getText"):
                text = child.getText().upper()
                if "IS" in text and "NULL" in text:
                    if "NOT" in text:
                        result = QueryExpressionFunction(
                            function=Function.IS_NOT_NULL,
                            parameters=[result] if isinstance(result, QueryExpression) else [],
                        )
                    else:
                        result = QueryExpressionFunction(
                            function=Function.IS_NULL,
                            parameters=[result] if isinstance(result, QueryExpression) else [],
                        )

        return result

    def visit_oC_PropertyOrLabelsExpression(self, ctx: Any) -> QueryExpression:
        """Visit a property or labels expression context."""
        result = self.visit(ctx.oC_Atom())

        # Handle property lookups
        for lookup_ctx in ctx.oC_PropertyLookup() or []:
            prop_name = lookup_ctx.oC_PropertyKeyName().getText()
            if isinstance(result, QueryExpressionProperty):
                result.property_name = prop_name
            elif isinstance(result, QueryExpression):
                result = QueryExpressionProperty(
                    variable_name=str(result),
                    property_name=prop_name,
                )

        return result

    # =========================================================================
    # Atom visitors
    # =========================================================================

    def visit_oC_Atom(self, ctx: Any) -> QueryExpression:
        """Visit an atom context.

        Grammar:
        oC_Atom
            :  oC_Literal
            | oC_Parameter
            | oC_CaseExpression
            | ( COUNT SP? '(' SP? '*' SP? ')' )
            | oC_ListComprehension
            | oC_PatternComprehension
            | oC_Quantifier
            | oC_Reduce
            | oC_PatternPredicate
            | oC_ParenthesizedExpression
            | oC_FunctionInvocation
            | oC_ExistentialSubquery
            | oC_Variable
            ;
        """
        if ctx.oC_Literal():
            return self.visit(ctx.oC_Literal())
        if ctx.oC_Parameter():
            return self.visit(ctx.oC_Parameter())
        if ctx.oC_Variable():
            return QueryExpressionProperty(variable_name=ctx.oC_Variable().getText())
        if ctx.oC_FunctionInvocation():
            return self.visit(ctx.oC_FunctionInvocation())
        if ctx.oC_ParenthesizedExpression():
            return self.visit(ctx.oC_ParenthesizedExpression())
        if ctx.oC_CaseExpression():
            return self.visit(ctx.oC_CaseExpression())
        if ctx.oC_ListComprehension():
            return self.visit(ctx.oC_ListComprehension())
        if ctx.oC_Quantifier():
            return self.visit(ctx.oC_Quantifier())
        if ctx.oC_Reduce():
            return self.visit(ctx.oC_Reduce())
        if ctx.oC_ExistentialSubquery():
            return self.visit(ctx.oC_ExistentialSubquery())
        if ctx.oC_PatternPredicate():
            return self.visit_oC_PatternPredicate(ctx.oC_PatternPredicate())
        # Check for COUNT(*) - it's a token, not a rule
        if hasattr(ctx, "COUNT") and ctx.COUNT():
            return QueryExpressionAggregationFunction(
                aggregation_function=try_parse_aggregation_function("count")
                or AggregationFunction.COUNT,
            )

        # Default: return a property with the text
        return QueryExpressionProperty(variable_name=ctx.getText())

    def visit_oC_Literal(self, ctx: Any) -> QueryExpressionValue:
        """Visit a literal context."""
        if ctx.oC_NumberLiteral():
            return self.visit(ctx.oC_NumberLiteral())
        if ctx.StringLiteral():
            text = ctx.StringLiteral().getText()
            # Remove quotes
            if (text.startswith("'") and text.endswith("'")) or (
                text.startswith('"') and text.endswith('"')
            ):
                text = text[1:-1]
            return QueryExpressionValue(value=text, value_type=str)
        if ctx.oC_BooleanLiteral():
            text = ctx.oC_BooleanLiteral().getText().lower()
            return QueryExpressionValue(value=text == "true", value_type=bool)
        if ctx.getText().upper() == "NULL":
            return QueryExpressionValue(value=None, value_type=type(None))
        if ctx.oC_ListLiteral():
            return self.visit(ctx.oC_ListLiteral())
        if ctx.oC_MapLiteral():
            return self.visit(ctx.oC_MapLiteral())

        return QueryExpressionValue(value=ctx.getText(), value_type=str)

    def visit_oC_NumberLiteral(self, ctx: Any) -> QueryExpressionValue:
        """Visit a number literal context."""
        if ctx.oC_IntegerLiteral():
            text = ctx.oC_IntegerLiteral().getText()
            value = int(text)
            return QueryExpressionValue(value=value, value_type=int)
        if ctx.oC_DoubleLiteral():
            text = ctx.oC_DoubleLiteral().getText()
            value = float(text)
            return QueryExpressionValue(value=value, value_type=float)

        return QueryExpressionValue(value=0, value_type=int)

    def visit_oC_ListLiteral(self, ctx: Any) -> QueryExpressionList:
        """Visit a list literal context."""
        items: list[QueryExpression] = []
        for expr_ctx in ctx.oC_Expression() or []:
            expr = self.visit(expr_ctx)
            if isinstance(expr, QueryExpression):
                items.append(expr)
        return QueryExpressionList(items=items)

    def visit_oC_MapLiteral(self, ctx: Any) -> QueryExpressionMapLiteral:
        """Visit a map literal context.

        Grammar:
        oC_MapLiteral
            :  '{' SP? ( oC_PropertyKeyName SP? ':' SP? oC_Expression SP?
               ( ',' SP? oC_PropertyKeyName SP? ':' SP? oC_Expression SP? )* )? '}' ;

        Examples:
        - {name: 'John', age: 30}
        - {year: 2024, month: 1, day: 15}
        - {}  (empty map)
        """
        entries: list[tuple[str, QueryExpression]] = []

        # Get all property key names and expressions
        key_ctxs = ctx.oC_PropertyKeyName() or []
        expr_ctxs = ctx.oC_Expression() or []

        # Pair them up
        for key_ctx, expr_ctx in zip(key_ctxs, expr_ctxs):
            key_name = key_ctx.getText()
            value_expr = self.visit(expr_ctx)
            if isinstance(value_expr, QueryExpression):
                entries.append((key_name, value_expr))

        return QueryExpressionMapLiteral(entries=entries)

    def visit_oC_FunctionInvocation(self, ctx: Any) -> QueryExpression:
        """Visit a function invocation context.

        Grammar:
        oC_FunctionInvocation
            :  oC_FunctionName SP? '(' SP? ( DISTINCT SP? )?
               ( oC_Expression SP? ( ',' SP? oC_Expression SP? )* )?
               ( SP? ORDER SP BY SP oC_SortItem ( SP? ',' SP? oC_SortItem )* )? ')' ;
        """
        func_name = ctx.oC_FunctionName().getText()

        # Check for DISTINCT
        is_distinct = any(
            ctx.getChild(i).getText().upper() == "DISTINCT"
            for i in range(ctx.getChildCount())
            if hasattr(ctx.getChild(i), "getText")
        )

        # Get parameters
        params: list[QueryExpression] = []
        for expr_ctx in ctx.oC_Expression() or []:
            expr = self.visit(expr_ctx)
            if isinstance(expr, QueryExpression):
                params.append(expr)

        # Parse ORDER BY for ordered aggregation (e.g., COLLECT(x ORDER BY y DESC))
        order_by: list[tuple[QueryExpression, bool]] = []
        sort_items = ctx.oC_SortItem() if hasattr(ctx, "oC_SortItem") else None
        if sort_items:
            for sort_item in sort_items if isinstance(sort_items, list) else [sort_items]:
                sort_expr = self.visit(sort_item.oC_Expression())
                is_desc = False
                # Check for DESC keyword
                if sort_item.DESC():
                    is_desc = True
                elif sort_item.DESCENDING():
                    is_desc = True
                if isinstance(sort_expr, QueryExpression):
                    order_by.append((sort_expr, is_desc))

        # Check if it's an aggregation function
        agg_func = try_parse_aggregation_function(func_name)
        if agg_func:
            return QueryExpressionAggregationFunction(
                aggregation_function=agg_func,
                is_distinct=is_distinct,
                inner_expression=params[0] if params else None,
                order_by=order_by,
            )

        # Check if it's a regular function
        func_info = try_get_function(func_name)
        if func_info:
            return QueryExpressionFunction(
                function=func_info.function_name,
                parameters=params,
            )

        # Unknown function - return as generic function
        return QueryExpressionFunction(
            function=Function.INVALID,
            parameters=params,
        )

    def visit_oC_ParenthesizedExpression(self, ctx: Any) -> QueryExpression:
        """Visit a parenthesized expression context."""
        return self.visit(ctx.oC_Expression())

    def visit_oC_CaseExpression(self, ctx: Any) -> QueryExpressionCaseExpression:
        """Visit a CASE expression context.

        Grammar:
        oC_CaseExpression:
            ( CASE ( oC_CaseAlternative )+ )                    # Simple CASE
            | ( CASE oC_Expression ( oC_CaseAlternative )+ )    # Searched CASE
            ( ELSE oC_Expression )? END

        oC_CaseAlternative: WHEN oC_Expression THEN oC_Expression
        """
        test_expr = None
        alternatives: list[tuple[QueryExpression, QueryExpression]] = []
        else_expr = None

        # Get all expressions - first one might be test expression
        exprs = ctx.oC_Expression() or []

        # Get all CASE alternatives (WHEN/THEN pairs)
        case_alternatives = ctx.oC_CaseAlternative() or []

        # If we have expressions AND case alternatives, first expression is test
        # Otherwise it's a simple CASE (no test expression)
        if exprs and case_alternatives:
            # Check if first expression is test expression or ELSE
            # ELSE expression comes after all alternatives, so if we have
            # more alternatives than expressions, first expr is test
            if len(exprs) == 1:
                # Only one expression - could be ELSE
                else_expr = self.visit(exprs[0])
            elif len(exprs) == 2:
                # Two expressions - first is test, second is ELSE
                test_expr = self.visit(exprs[0])
                else_expr = self.visit(exprs[1])
            else:
                # Multiple - first is test, last is ELSE
                test_expr = self.visit(exprs[0])
                else_expr = self.visit(exprs[-1])
        elif exprs and not case_alternatives:
            # No alternatives - this shouldn't happen in valid CASE
            pass

        # Process WHEN/THEN pairs from oC_CaseAlternative
        for alt in case_alternatives:
            alt_exprs = alt.oC_Expression() or []
            if len(alt_exprs) >= 2:
                when_expr = self.visit(alt_exprs[0])
                then_expr = self.visit(alt_exprs[1])
                if isinstance(when_expr, QueryExpression) and isinstance(
                    then_expr, QueryExpression
                ):
                    alternatives.append((when_expr, then_expr))

        return QueryExpressionCaseExpression(
            test_expression=test_expr,
            alternatives=alternatives,
            else_expression=else_expr,
        )

    def visit_oC_ExistentialSubquery(self, ctx: Any) -> QueryExpressionExists:
        """Visit an existential subquery context.

        Grammar:
        oC_ExistentialSubquery
            :  EXISTS SP? '{' SP? ( oC_RegularQuery | ( oC_Pattern ( SP? oC_Where )? ) ) SP? '}' ;

        This represents EXISTS { pattern } or EXISTS { MATCH ... RETURN ... } constructs.
        """
        pattern_entities: list[Entity] = []
        where_expression: QueryExpression | None = None
        subquery: QueryNode | None = None

        # Check for full subquery (oC_RegularQuery)
        if ctx.oC_RegularQuery():
            subquery = self.visit(ctx.oC_RegularQuery())
        # Check for pattern-based EXISTS
        elif ctx.oC_Pattern():
            pattern_entities = self.visit(ctx.oC_Pattern())
            if isinstance(pattern_entities, Entity):
                pattern_entities = [pattern_entities]
            elif not isinstance(pattern_entities, list):
                pattern_entities = []

            # Check for optional WHERE clause
            if ctx.oC_Where():
                where_result = self.visit(ctx.oC_Where())
                if isinstance(where_result, QueryExpression):
                    where_expression = where_result

        return QueryExpressionExists(
            pattern_entities=pattern_entities,
            where_expression=where_expression,
            is_negated=False,  # NOT is handled in expression visitor
            subquery=subquery,
        )

    def visit_oC_PatternPredicate(self, ctx: Any) -> QueryExpressionExists:
        """Visit a pattern predicate context.

        Grammar:
        oC_PatternPredicate: oC_RelationshipsPattern ;

        In OpenCypher, a pattern in a boolean context (e.g., WHERE clause) represents
        an implicit EXISTS check. For example:
            WHERE (c)-[:HAS_LOAN]->(:Loan)
        is equivalent to:
            WHERE EXISTS { (c)-[:HAS_LOAN]->(:Loan) }

        This method converts the pattern into a QueryExpressionExists node.
        """
        pattern_entities: list[Entity] = []

        # Visit the relationships pattern to extract entities
        # oC_RelationshipsPattern: oC_NodePattern ( SP? oC_PatternElementChain )+
        if ctx.oC_RelationshipsPattern():
            rel_ctx = ctx.oC_RelationshipsPattern()

            # Get the starting node pattern
            if rel_ctx.oC_NodePattern():
                node = self.visit(rel_ctx.oC_NodePattern())
                if isinstance(node, NodeEntity):
                    pattern_entities.append(node)

            # Get the chain elements (relationships and nodes)
            for chain_ctx in rel_ctx.oC_PatternElementChain() or []:
                chain = self.visit(chain_ctx)
                if isinstance(chain, list):
                    pattern_entities.extend(chain)
                elif isinstance(chain, Entity):
                    pattern_entities.append(chain)

        # Create an EXISTS expression with the pattern
        # The NOT operator (if present) is handled by the parent expression visitor
        return QueryExpressionExists(
            pattern_entities=pattern_entities,
            where_expression=None,
            is_negated=False,
            subquery=None,
        )

    def visit_oC_Parameter(self, ctx: Any) -> QueryExpressionParameter:
        """Visit a parameter context.

        Grammar:
        oC_Parameter: '$' ( oC_SymbolicName | DecimalInteger ) ;

        Handles parameter syntax like $watchlist, $1, etc.
        """
        # Get the full text and remove the '$' prefix
        text = ctx.getText()
        param_name = text[1:] if text.startswith("$") else text
        return QueryExpressionParameter(parameter_name=param_name)

    def visit_oC_Quantifier(self, ctx: Any) -> QueryExpressionListPredicate:
        """Visit a quantifier expression context.

        Grammar:
        oC_Quantifier: ( ALL SP? '(' SP? oC_FilterExpression SP? ')' )
                     | ( ANY SP? '(' SP? oC_FilterExpression SP? ')' )
                     | ( NONE SP? '(' SP? oC_FilterExpression SP? ')' )
                     | ( SINGLE SP? '(' SP? oC_FilterExpression SP? ')' );

        oC_FilterExpression: oC_IdInColl ( SP? oC_Where )? ;
        oC_IdInColl: oC_Variable SP IN SP oC_Expression ;

        Examples:
        - ALL(x IN list WHERE x > 0)
        - ANY(x IN accounts WHERE x.balance < 0)
        - NONE(x IN transactions WHERE x.suspicious)
        - SINGLE(x IN matches WHERE x.valid)
        """
        # Determine predicate type from the keyword
        text = ctx.getText().upper()
        if text.startswith("ALL"):
            predicate_type = ListPredicateType.ALL
        elif text.startswith("ANY"):
            predicate_type = ListPredicateType.ANY
        elif text.startswith("NONE"):
            predicate_type = ListPredicateType.NONE
        elif text.startswith("SINGLE"):
            predicate_type = ListPredicateType.SINGLE
        else:
            predicate_type = ListPredicateType.ANY  # fallback

        # Get the filter expression
        filter_expr_ctx = ctx.oC_FilterExpression()
        if not filter_expr_ctx:
            raise TranspilerSyntaxErrorException(
                "Quantifier requires a filter expression"
            )

        # Parse oC_IdInColl: oC_Variable SP IN SP oC_Expression
        id_in_coll_ctx = filter_expr_ctx.oC_IdInColl()
        if not id_in_coll_ctx:
            raise TranspilerSyntaxErrorException(
                "Quantifier requires variable IN list syntax"
            )

        variable_name = id_in_coll_ctx.oC_Variable().getText()
        list_expression = self.visit(id_in_coll_ctx.oC_Expression())

        # Parse optional WHERE clause
        filter_expression: QueryExpression | None = None
        where_ctx = filter_expr_ctx.oC_Where()
        if where_ctx:
            filter_expression = self.visit(where_ctx)

        return QueryExpressionListPredicate(
            predicate_type=predicate_type,
            variable_name=variable_name,
            list_expression=list_expression,
            filter_expression=filter_expression,
        )

    def visit_oC_ListComprehension(self, ctx: Any) -> QueryExpressionListComprehension:
        """Visit a list comprehension context.

        Grammar:
        oC_ListComprehension
            :  '[' SP? oC_FilterExpression ( SP? '|' SP? oC_Expression )? SP? ']' ;

        oC_FilterExpression: oC_IdInColl ( SP? oC_Where )? ;
        oC_IdInColl: oC_Variable SP IN SP oC_Expression ;

        Examples:
        - [x IN list] -> just the list (identity)
        - [x IN list WHERE x > 0] -> FILTER(list, x -> x > 0)
        - [x IN list | x * 2] -> TRANSFORM(list, x -> x * 2)
        - [x IN list WHERE x > 0 | x * 2] -> TRANSFORM(FILTER(list, x -> x > 0), x -> x * 2)
        """
        # Get the filter expression (contains variable IN list WHERE pred)
        filter_expr_ctx = ctx.oC_FilterExpression()
        if not filter_expr_ctx:
            raise TranspilerSyntaxErrorException(
                "List comprehension requires a filter expression"
            )

        # Parse oC_IdInColl: oC_Variable SP IN SP oC_Expression
        id_in_coll_ctx = filter_expr_ctx.oC_IdInColl()
        if not id_in_coll_ctx:
            raise TranspilerSyntaxErrorException(
                "List comprehension requires variable IN list syntax"
            )

        variable_name = id_in_coll_ctx.oC_Variable().getText()
        list_expression = self.visit(id_in_coll_ctx.oC_Expression())

        # Parse optional WHERE clause (filter)
        filter_expression: QueryExpression | None = None
        where_ctx = filter_expr_ctx.oC_Where()
        if where_ctx:
            filter_expression = self.visit(where_ctx)

        # Parse optional map expression (after |)
        map_expression: QueryExpression | None = None
        # The map expression is the oC_Expression directly under oC_ListComprehension
        # (not the one inside oC_FilterExpression)
        exprs = ctx.oC_Expression()
        if exprs:
            # The grammar has oC_Expression after the |
            map_expression = self.visit(exprs)

        return QueryExpressionListComprehension(
            variable_name=variable_name,
            list_expression=list_expression,
            filter_expression=filter_expression,
            map_expression=map_expression,
        )

    def visit_oC_Reduce(self, ctx: Any) -> QueryExpressionReduce:
        """Visit a REDUCE expression context.

        Grammar:
        oC_Reduce
            :  REDUCE SP? '(' SP? oC_Variable SP? '=' SP? oC_Expression SP? ','
               SP? oC_IdInColl SP? '|' SP? oC_Expression SP? ')' ;

        oC_IdInColl: oC_Variable SP IN SP oC_Expression ;

        Example:
        REDUCE(total = 0, x IN amounts | total + x)
        -> AGGREGATE(amounts, 0, (total, x) -> total + x)
        """
        # Get accumulator variable (the first oC_Variable in the rule)
        acc_var_ctx = ctx.oC_Variable()
        if not acc_var_ctx:
            raise TranspilerSyntaxErrorException(
                "REDUCE requires an accumulator variable"
            )
        accumulator_name = acc_var_ctx.getText()

        # Get all expressions - there are 2 in the main rule:
        # 1. Initial value (after =)
        # 2. Reducer expression (after |)
        # The list expression is inside oC_IdInColl
        expressions = ctx.oC_Expression()
        if not expressions or len(expressions) < 2:
            raise TranspilerSyntaxErrorException(
                "REDUCE requires initial value and reducer expression"
            )

        initial_value = self.visit(expressions[0])
        reducer_expression = self.visit(expressions[1])

        # Get oC_IdInColl for variable IN list
        id_in_coll_ctx = ctx.oC_IdInColl()
        if not id_in_coll_ctx:
            raise TranspilerSyntaxErrorException(
                "REDUCE requires variable IN list syntax"
            )

        variable_name = id_in_coll_ctx.oC_Variable().getText()
        list_expression = self.visit(id_in_coll_ctx.oC_Expression())

        return QueryExpressionReduce(
            accumulator_name=accumulator_name,
            initial_value=initial_value,
            variable_name=variable_name,
            list_expression=list_expression,
            reducer_expression=reducer_expression,
        )

    # =========================================================================
    # Helper methods
    # =========================================================================

    def _visit_binary_expression(
        self,
        ctx: Any,
        sub_expressions: list[Any],
        operator_keyword: str,
    ) -> QueryExpression:
        """Visit a binary expression with a single operator type."""
        if not sub_expressions:
            return QueryExpressionValue(value=None, value_type=type(None))

        result = self.visit(sub_expressions[0])

        if len(sub_expressions) > 1:
            op_info = try_get_operator(operator_keyword)
            for sub_expr in sub_expressions[1:]:
                right = self.visit(sub_expr)
                if isinstance(right, QueryExpression) and isinstance(result, QueryExpression):
                    result = QueryExpressionBinary(
                        operator=op_info,
                        left_expression=result,
                        right_expression=right,
                    )

        return result

    def _visit_chained_binary_expression(
        self,
        ctx: Any,
        sub_expressions: list[Any],
        operators: list[str],
    ) -> QueryExpression:
        """Visit a chained binary expression with multiple operator types."""
        if not sub_expressions:
            return QueryExpressionValue(value=None, value_type=type(None))

        result = self.visit(sub_expressions[0])

        # Find operators in the children
        op_index = 0
        for i in range(ctx.getChildCount()):
            child = ctx.getChild(i)
            if not hasattr(child, "getRuleIndex"):
                text = child.getText()
                if text in operators and op_index < len(sub_expressions) - 1:
                    op_index += 1
                    right = self.visit(sub_expressions[op_index])
                    op_info = try_get_operator(text)
                    if isinstance(right, QueryExpression) and isinstance(result, QueryExpression):
                        result = QueryExpressionBinary(
                            operator=op_info,
                            left_expression=result,
                            right_expression=right,
                        )

        # If we didn't find operators in children, just visit the rest
        if op_index == 0:
            for i, sub_expr in enumerate(sub_expressions[1:], 1):
                right = self.visit(sub_expr)
                if isinstance(right, QueryExpression) and isinstance(result, QueryExpression):
                    result = QueryExpressionBinary(
                        operator=try_get_operator(operators[0] if operators else "+"),
                        left_expression=result,
                        right_expression=right,
                    )

        return result
